
/*
 *  CPlan solves planning problems formulated as constraint satisfaction
 *  problems.
 *
 *  Copyright (C) 1999  Peter van Beek and Xinguang Chen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.  See the GNU General Public License
 *  for more details (see the file called Copying or contact the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA).
 */

#include "csp.h"
#include "stack.h"
#include "global.h"

/*
 * Three global variables to count properties discovered by the backtrack
 * routine during its execution: "checks" is the number of consistency checks
 * performed by the routine, "solution_count" is the number of solutions
 * found by the routine, and "visits" is the number of nodes of the search
 * tree visited by the routine.
 */
double  solution_count;
double  checks;
double  visits;

/*
 *  The conflicts information is divided into two pieces.
 *  
 *  1. The conflicts found in the backward phase (those from the conflict set
 *     of a deeper variable in a backjumping) are recorded in conflicts.
 *     conflicts[x][y] = 1 means y is in the conflict set of x.
 *  
 *  2. The conflicts found in the forward phase are recorded in checking.
 *     checking[x][y] == i means that the instantiation of past variable x was
 *     first checked against variable y at level i and caused the domain of y
 *     to be pruned. Only the shallowest level is kept. Thus, until a
 *     backtracking occurs to level i, x is still in the conflict set of y.
 *  
 *  When a dead-end is encountered, these two pieces of information are merged
 *  together to form a complete conflict set for the current variable.
 */
int     **checking;
BOOL    **conflicts;

/*
 *  The overhead in the procedure record_checking which records the conflicts
 *  information generated during constraint propagation and the procedure
 *  restore whichs restores the domains and conflicts information in a
 *  backtracking step can be reduced by using two auxiliary data-structures,
 *  checking_set and checking_top.  The data-structure checking implements
 *  a table representation of the conflicts information founded in
 *  the forward phase of the backtrack search, while checking_set and
 *  checking_top implement a list representation of the above conflicts
 *  information, in which checking_top[y] denotes the number of the past
 *  variables x such that checking[x][y] > 0 and checking_set[y][1], ...,
 *  checking_set[y][checking_top[y]] record each of these variables.
 */
int	**checking_set;
int	*checking_top;

/*
 *  domains[x][a] denotes whether value a is active in the domain of variable
 *  x, where domain[x][a] = 0 indicates that a is still an active value in its
 *  domain, and domain[x][a] = i > 0 indicates that a has been removed from the
 *  domain at level i of the backtrack search.
 */
int     **domains;

/*
 *  count_uninst[C] counts the number of uninstantiated variables in the
 *  scheme of constraint C; it is initialized to be |vars(C)|. It is used
 *  to control whether the constraint C is checkable at a particular stage
 *  in the search.
 */
int     *count_uninst;

/*
 *  instantiated[x] marks whether variable x is currently instantiated;
 *  initially it is set to be false.
 */
BOOL    *instantiated;

/*
 *  order[i] identifies the i'th variable instantiated in the current
 *  partial solution.
 */
int     *order;

/*
 *  domain_count[x] records the number of active values in the domain of
 *  variable x and degree[x] records the degree of variable x. They are used
 *  in the dynamic variable ordering routine.
 */
int     *domain_count;
int	*degree;

/*
 *  Dynamic variable ordering routine.
 *
 *  Choose the next variable to be instantiated by searching for the
 *  variable with the least number of values still available.
 *  Break ties by choosing the variable that participates in the most
 *  constraints (statically determined before problem-solving starts).
 *  Hidden variables always come last.
 */
static int
get_next_var( C, level )
    CSP_type    *C;
    int         level;
{
    int i, j, var, min_var;
    /*int min_count;*/
    double x, min_count;

    /*
     *  Record the number of constraints each variable
     *  participates in. Used to break ties.
     */
    if( level == 1 ) {
        for( i = 0; i < C->m; i++ ) {
            for( j = 0; j < C->arity[i]; j++ ) {
                var = C->scheme[i][j];
                degree[var]++;
            }
        }
    }

    /*
     *  All visible variables have been instantiated.
     *  Return first uninstantiated hidden variable.
     */
    if( level > C->n_visible ) {
        for( i = 1; i <= C->n; i++ ) if( !instantiated[i] ) {
            return( i );
        }
        return( 0 ); /* something has gone wrong */
    }
    /*
     *  There are uninstantiated visible variables.
     *  Find visible variable with min domain.
     */
    else {
        min_count = HUGE; /* larger than any possible domain size */
        min_var = 0;      /* guaranteed to be changed             */
        for( i = 1; i <= C->n; i++ ) {
            if( !instantiated[i] &&
                        C->visible[C->variable_attribute[i].type] ) {

/*
begin: dom+deg
                if( domain_count[i] < min_count ) {
                    min_count = domain_count[i];
                    min_var = i;
                }
                else
                if( (domain_count[i] == min_count) &&
                    (degree[i] > degree[min_var]) ) {
                    min_var = i;
                }
                else
                if( (domain_count[i] == min_count) &&
                    (degree[i] == degree[min_var]) &&
                    (C->variable_attribute[i].state >
                        C->variable_attribute[min_var].state) ) {
                    min_var = i;
                }
end: dom+deg
*/

/*
begin: dom/deg
*/
	        x = (double)domain_count[i] / (double)degree[i];
	        if( x < min_count ) {
	            min_count = x;
	            min_var = i;
	        }
/*
end: dom/deg
*/

            }
        }

        return( min_var );
    }
}

/*
 *  Restore the domains and conflict information because of a backtracking
 *  step to current_level. A variable is known to have had a value deleted
 *  by looking at the "checking" matrix.
 */
static void
restore( C, current_level, current_variable )
    CSP_type    *C;
    int         current_level, current_variable;
{
    int         j, a;

    for( j = 1; j <= C->n; j++ ) if( !instantiated[j] ) {
	if( checking[current_variable][j] == current_level ) {
	    for( a = 0; a < C->domain_size[j]; a++ ) {
		if( domains[j][a] == current_level ) {
		    domains[j][a] = 0;
		    domain_count[j]++;
		}
	    }
	    while( checking_top[j] && 
	    	   checking[checking_set[j][checking_top[j]]][j] == current_level ) {
	        checking[checking_set[j][checking_top[j]]][j] = 0;
		checking_top[j]--;
	    }
	}
    }
}

/*
 *  As "current_variable" has been chosen to be instantiated next,
 *  update (decrease) the number of uninstantiated variables for
 *  each of the constraints involving "current_variable".
 */
static void
update_constraint_counts( C, current_variable )
    CSP_type    *C;
    int         current_variable;
{
    int         i, r;

    for( i = 0; i < C->n_cons[current_variable]; i++ ) {
        r = C->index[current_variable][i];
        count_uninst[r]--;
    }
}

/*
 *  As backtracking has occurred at the "current_level",
 *  restore (increase) the number of uninstantiated variables for
 *  each of the constraints involving "current_variable".
 */
static void
restore_constraint_counts( C, current_variable )
    CSP_type    *C;
    int         current_variable;
{
    int         i, r;

    for( i = 0; i < C->n_cons[current_variable]; i++ ) {
        r = C->index[current_variable][i];
        count_uninst[r]++;
    }
}

static void
union_conflicts( int x, int y )
{
  int i;
  for( i = 1; order[i] != x; i++ )
     conflicts[x][order[i]] =
         conflicts[x][order[i]] ||
         conflicts[y][order[i]];
}

static void
empty_conflicts( int from, int to )
{
  int i, j;
  for( i = from+1; i <= to; i++ )
      for( j = 1; j <= i; j++ )
          conflicts[order[i]][order[j]] = 0;
}

static void
union_checking( int x, int y )
{
  int i;
  for( i = 1; order[i] != x; i++ )
      conflicts[x][order[i]] =
          conflicts[x][order[i]] ||
          (checking[order[i]][y] != 0);
}

/*
 *  Record the conflict information generated during constraint propagation.
 *
 *  Whenever value a of a future variable x fails to find a valid support in a
 *  constraint and thus is removed from its domain, GAC-CBJ needs to compute
 *  a no-good accountable for such a removal. The failure of x <-- a to find
 *  a valid support in a constraint C may be due to:
 *
 *  1. an instantiation of a past variable does not support x <-- a in C and
 *     thus the past variable should be added to the current conflict set of
 *     variable x.
 *
 *  2. a value of an uninstantiated variable var, which could be used to form
 *     a support for a in C, is removed from its domain. Thus the current
 *     conflict set of var, which is accountable for current removings in the
 *     domain of var, should be merged in the current conflict set of x.
 *
 *  Procedure record_checking in GAC-CBJ records all the above conflicts
 *  information in the checking data structure for x.
 */
static void
record_checking( C, r, x, current_level )
    CSP_type	*C;
    int		r;
    int		x;
    int		current_level;
{
    int		i, j, var, top;

    for( i = 0; i < C->arity[r]; i++ ) {
	var = C->scheme[r][i];
	if( var != x ) {
	    if( instantiated[var] ) {
	    	if( checking[var][x] == 0 ) {
		    top = ++checking_top[x];
		    checking[var][x] = current_level;
		    checking_set[x][top] = var;
		}
	    }
	    else {
	        for( j = 1; j <= checking_top[var]; j++ ) {
		    if( checking[checking_set[var][j]][x] == 0 ) {
		        top = ++checking_top[x];
			checking[checking_set[var][j]][x] = current_level;
			checking_set[x][top] = checking_set[var][j];
		    }
		}
	    }
	}
    }
}

/*
 *  A constraint is forward checkable if one of its
 *  variables is uninstantiated.
 */
static int
forward_checkable( C, r )
    CSP_type    *C;
    int         r;
{
    return( count_uninst[r] == 1 );
}

/*
 *  A constraint is arc consistency checkable if two or
 *  more of its variables are uninstantiated. (If only
 *  one of the variables is uninstantiated it is not
 *  arc consistency checkable since the constraint has
 *  already been forward checked.)
 */
static int
arc_consistency_checkable( C, r )
    CSP_type    *C;
    int         r;
{
    return( 2 <= count_uninst[r] &&
		 count_uninst[r] <= C->checkable[C->relation[r]] );
}

/*
 *  Check a ``forward_checkable'' constraint to see which values in
 *  the domain of a future variable can be elminated due to the
 *  instantiations so far.
 */
static int
check_forward( C, r, current_level, current_variable, solution, fail )
    CSP_type    *C;
    int         r, current_level, current_variable;
    int         *solution;
    int         *fail;
{
    int i, a, future_variable;
    int delete_count = 0;

    /*
     *  Which is the uninstantiated variable?
     */
    for( i = 0; i < C->arity[r] &&
                instantiated[C->scheme[r][i]];
         i++ ) {
            ;
    }
    future_variable = C->scheme[r][i];

    for( a = 0; a < C->domain_size[future_variable]; a++ )
        if( domains[future_variable][a] == 0 ) {

            /* temporarily instantiate to check */
            solution[future_variable] = a;
            if( !check_constraint( C, r, solution ) ) {
                domains[future_variable][a] = current_level;
                domain_count[future_variable]--;
                delete_count++;
            }
        }

    if( delete_count ) {
        if( domain_count[future_variable] ) /* domain not empty */
            push( future_variable );
        record_checking(C, r, future_variable, current_level);
    }

    *fail = future_variable;
    return( domain_count[future_variable] );
}

/*
 *  Given instantiations of some of the variables, does there exist
 *  values for the remaining variables in the scheme of the relation
 *  such that the constraint is satisfied?
 */
static int
exists( C, rel_index, solution )
    CSP_type    *C;
    int         rel_index;
    int         *solution;
{
    int         i, var, b;
    int         found;

    int         tuple[C->arity[rel_index]];
    int         min_domain[C->arity[rel_index]];

    /*
     *  Initialize the starting tuple.
     */
    for( i = 0; i < C->arity[rel_index]; i++ ) {
        var = C->scheme[rel_index][i];
        if( instantiated[var] )
            tuple[i] = solution[var];
        else {
            /* find first valid element */
            found = 0;
            for( b = 0; b < C->domain_size[var] && !found; b++ ) {
                if( domains[var][b] == 0 ) {
                    tuple[i] = min_domain[i] = b;
                    found = 1;
                    break;
                }
            }
            if( !found ) {
                return( 0 );
            }
        }
    }

    i = C->arity[rel_index]-1;
    while( i >= 0 ) {

        for( i = 0; i < C->arity[rel_index]; i++ ) {
            var = C->scheme[rel_index][i];
            solution[var] = tuple[i];
        }

        if( check_constraint( C, rel_index, solution ) ) {
            return( 1 );
        }

        /* get next tuple */

        for( i = C->arity[rel_index]-1; i >= 0; i-- ) {
            var = C->scheme[rel_index][i];
            if( !instantiated[var] ) {
                found = 0;
                var = C->scheme[rel_index][i];
                for( b = tuple[i]+1; b < C->domain_size[var]; b++ )
                    if( domains[var][b] == 0) {
                        found = 1;
                        tuple[i] = b;
                        break;
                    }
                if( found) break;
                tuple[i] = min_domain[i];
            }
        }
    }

    return( 0 );
}

/*
 *  The domain of variable v1 is to be filtered based on the given constraint.
 */
static int
revise( C, rel_index, v1, current_level, current_variable, solution, fail )
    CSP_type    *C;
    int         rel_index;
    int         v1, current_level, current_variable;
    int         *solution;
    int         *fail;
{
    int a, changed = 0, is_consistent = 0;

    /*
     *  For each element in the domain of v1, check if there
     *  exists supporting elements in the domains of the remaining
     *  variables in the scheme of the relation.
     */
    instantiated[v1] = 1;
    for( a = 0; a < C->domain_size[v1]; a++ )
        if( domains[v1][a] == 0 ) {
            solution[v1] = a;
            if( exists( C, rel_index, solution ) ) {
                is_consistent = 1;
            }
            else {
                changed = 1;
                domains[v1][a] = current_level;
                domain_count[v1]--;
            }
        }
    instantiated[v1] = 0;

    if( changed ) {
        push( v1 );
        record_checking(C, rel_index, v1, current_level);
    }

    *fail = v1;
    return( is_consistent );
}

static int
consistent( C, current_level, current_variable, solution, fail )
    CSP_type    *C;
    int         current_level, current_variable;
    int         *solution;
    int         *fail;
{
    int         i, j, rel_index, v;
    /*
     *  for each constraint in which current_variable participates...
     */
    for( j = 0; j < C->n_cons[current_variable]; j++ ) {
        rel_index = C->index[current_variable][j];
        if( forward_checkable( C, rel_index ) ) {
            if( !check_forward( C, rel_index,
                                current_level,
                                current_variable,
                                solution,
                                fail ) )
                goto END;
        }
    }

    push( current_variable );

    while( !stack_empty() ) {
        v = pop();

        /*
         *  Variable `v' has changed so need to revise any uninstantiated
         *  variables which share a constraint with `v'.
         *
         *  for each constraint in which the variable `v' participates...
         */
        for( j = 0; j < C->n_cons[v]; j++ ) {
            rel_index = C->index[v][j];
            if( arc_consistency_checkable( C, rel_index ) ) {

                /*
                 *  for each uninstantiated variable in the scheme...
                 */
                for( i = 0; i < C->arity[rel_index]; i++ ) {
                    if( C->scheme[rel_index][i] != v &&
                        !instantiated[C->scheme[rel_index][i]] ) {

                        if( !revise( C, rel_index,
                                     C->scheme[rel_index][i],
                                     current_level,
                                     current_variable,
                                     solution,
                                     fail ) )
                            goto END;

                    }
                }

            }
        }
    }
    return( 1 );

    END: while( !stack_empty() )
        v = pop();
    return( 0 );
}

/*
 *  Solve the constraint network by backtracking search interleaved
 *  with constraint propagation and conflict-based backjumping.
 */
static int
GAC_CBJ( C, solution, current_level )
    CSP_type    *C;
    int         current_level;
    int         *solution;
{
    int         i, a, current_variable;
    int         fail, jump;
    int         saved_solution_count = solution_count;

    if( current_level > C->n ) {
        solution_count++;
        process_solution( C, solution );
        return 0;
    }

    if( limit_exceeded() ) {
        return( 1 );
    }

    current_variable = get_next_var( C, current_level );
    order[current_level] = current_variable;

    update_constraint_counts( C, current_variable );

    for( a = 0; a < C->domain_size[current_variable]; a++ ) {
        if( domains[current_variable][a] == 0 ) {

            solution[current_variable] = a;
            instantiated[current_variable] = 1;
            visits++;

            if( consistent( C, current_level, current_variable, solution, &fail ) ) {
               jump = GAC_CBJ( C, solution, current_level+1 );
               if( jump != current_level ) {
                  restore( C, current_level, current_variable);
                  instantiated[current_variable] = 0;
                  restore_constraint_counts(C, current_variable);
                  return (jump);
                  }
               }
            else {
               union_checking(current_variable, fail);
               }
            restore( C, current_level, current_variable );
            instantiated[current_variable] = 0;
        }
    }

    union_checking(current_variable, current_variable);
    if( saved_solution_count == solution_count ) {
       jump = 0;
       for( i = current_level-1; i >= 1 && jump == 0; i-- )
           if( conflicts[current_variable][order[i]] )
              jump = i;
       }
    else {
       jump = current_level - 1;
    }

    if( jump >= 1 ) {
       union_conflicts( order[jump], current_variable);
       empty_conflicts( jump, current_level );
       }

    restore_constraint_counts( C, current_variable );
    return jump;
}

/*
 *  Interface to backtracking routine.
 */
void
solve( C )
    CSP_type    *C;
{
    int         i, j;
    int         *solution;

    /*
     *  Stack for arc consistency.
     */
    create_stack( C->n+1 );

    solution = new_calloc( C->n+1, sizeof(int) );
    domains = (int **)new_calloc( C->n+1, sizeof(int *) );
    domain_count = (int *)new_calloc( C->n+1, sizeof(int));
    for( i = 1; i <= C->n; i++ ) {
        domains[i] = new_calloc( C->domain_size[i], sizeof(int) );
        for( j = 0; j < C->domain_size[i]; j++ )
            domains[i][j] = 0;
        domain_count[i] = C->domain_size[i];
    }

    checking     = (int **)new_calloc( C->n+1, sizeof(int *) );
    checking_set = (int **)new_calloc( C->n+1, sizeof(int *) );
    checking_top = (int *)new_calloc( C->n+1, sizeof(int) );
    for( i = 1; i <= C->n; i++ ) {
        checking[i]     = (int *)new_calloc( C->n+1, sizeof(int) );
        checking_set[i] = (int *)new_calloc( C->n+1, sizeof(int) );
	checking_top[i] = 0;
        for( j = 1; j <= C->n; j++ )
            checking[i][j] = 0;
    }

    conflicts = (BOOL **)new_calloc( C->n+1, sizeof(BOOL *));
    for( i = 1; i <= C->n; i++ ) {
         conflicts[i] = (BOOL *)new_calloc( C->n+1, sizeof(BOOL));
         for( j = 1; j <= C->n; j++ )
             conflicts[i][j] = 0;
    }

    instantiated = (BOOL *)new_calloc( C->n+1, sizeof(BOOL) );
    order = new_calloc( C->n+1, sizeof(int));
    count_uninst = new_calloc( C->m, sizeof(int) );
    for( i = 0; i < C->m; i++ ) {
        count_uninst[i] = C->arity[i];
    }
    degree = new_calloc( C->n+1, sizeof(int) );

    checks = 0;
    solution_count = 0;
    visits = 0;
    GAC_CBJ( C, solution, 1 );

    destroy_stack();

    free( solution );

    for( i = 1; i <= C->n; i++ ) {
        free( domains[i] );
    }
    free( domains );

    for( i = 1; i <= C->n; i++ ) {
        free( checking[i] );
	free( checking_set[i] );
    }
    free( checking );
    free( checking_set );
    free( checking_top );

    for( i = 1; i <= C->n; i++ ) {
        free( conflicts[i] );
    }
    free( conflicts );

    free( instantiated );
    free( order );
    free( domain_count );
    free( count_uninst );
    free( degree );
}

