
/*
 *  CPlan solves planning problems formulated as constraint satisfaction
 *  problems.
 *
 *  Copyright (C) 1999  Peter van Beek and Xinguang Chen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.  See the GNU General Public License
 *  for more details (see the file called Copying or contact the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA).
 */

#include "grid.h"
#include "../Share/csp.h"
#include "../Share/global.h"


static int      check_key_distance_constraint();
static int      check_key_distance2_constraint();
static int      check_key_distance3_constraint();
static int      check_robot_distance_constraint();
static int      check_robot_distance2_constraint();
static int      check_robot_distance3_constraint();
static int      check_action_move_constraint();
static int      check_action_pickup_constraint();
static int      check_action_pickup_loose_constraint();
static int      check_action_putdown_constraint();
static int      check_action_unlock_constraint();
static int      check_useless_action_constraint();
static int      check_state_constraint();
static int      check_detent_constraint();
static int      check_noentry_constraint();

/*
 *  Check the given values against the constraint.
 */
int
check_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    checks += C->arity[r];

    switch( C->relation[r] ) {

    case KEY_DISTANCE_CONSTRAINT:
        return( check_key_distance_constraint( C, r, solution ) );

    case KEY_DISTANCE2_CONSTRAINT:
        return( check_key_distance2_constraint( C, r, solution ) );

    case KEY_DISTANCE3_CONSTRAINT:
        return( check_key_distance3_constraint( C, r, solution ) );

    case ROBOT_DISTANCE_CONSTRAINT:
        return( check_robot_distance_constraint( C, r, solution ) );

    case ROBOT_DISTANCE2_CONSTRAINT:
        return( check_robot_distance2_constraint( C, r, solution ) );

    case ROBOT_DISTANCE3_CONSTRAINT:
        return( check_robot_distance3_constraint( C, r, solution ) );

    case ACTION_MOVE_CONSTRAINT:
        return( check_action_move_constraint( C, r, solution ) );

    case ACTION_PICKUP_CONSTRAINT:
        return( check_action_pickup_constraint( C, r, solution ) );

    case ACTION_PICKUP_LOOSE_CONSTRAINT:
        return( check_action_pickup_loose_constraint( C, r, solution ) );

    case ACTION_PUTDOWN_CONSTRAINT:
        return( check_action_putdown_constraint( C, r, solution ) );

    case ACTION_UNLOCK_CONSTRAINT:
        return( check_action_unlock_constraint( C, r, solution ) );

    case USELESS_ACTION_CONSTRAINT:
        return( check_useless_action_constraint( C, r, solution ) );

    case STATE_CONSTRAINT:
        return( check_state_constraint( C, r, solution ) );

    case DETENT_CONSTRAINT:
        return( check_detent_constraint( C, r, solution ) );

    case NOENTRY_CONSTRAINT:
        return( check_noentry_constraint( C, r, solution ) );

    default:
        printf("Error in check_constraint.\n");
        exit(6);
    }

    return( 1 );
}

static int
same_attribute( attr1, attr2 )
    VALA        attr1, attr2;
{
    return( attr1.type   == attr2.type &&
            attr1.number == attr2.number );
}

/*
 *  Determine a lower bound on the number of time steps that would
 *  be needed to change the value of a key variable.
 */
static int
key_distance( a, a_new )
    VALA        a, a_new;
{
    if( a.type == GRID && a_new.type == GRID ) {
        if( a.number == a_new.number ) return( 0 );
        else return( grid_distance( a.number, a_new.number ) + 2 );
    }
    else
    if( ( a.type == GRID && a_new.type == HELD ) ||
        ( a.type == HELD && a_new.type == GRID ) ) {
        return( 1 );
    }
    else {
        return( 0 );
    }
}

/*
 *  Determine a lower bound on the number of time steps that would
 *  be needed to change the values of a pair of key variables.
 */
static int
key_distance2( a, a_new, b, b_new )
    VALA        a, a_new, b, b_new;
{
    int         dis, dis1, dis2, dis3;

    if( a.type == GRID && a_new.type == GRID ) {
        if( a.number == a_new.number ) return( 0 );
        if( b.type == GRID && b_new.type == GRID ) {

            if( b.number == b_new.number ) return( 0 );

	    dis  = grid_distance( a.number, a_new.number ) + 2 +
	    	   grid_distance( a_new.number, b.number ) +
		   grid_distance( b.number, b_new.number ) + 2;
	    if( a_new.number == b.number ) dis--;

	    dis1 = grid_distance( a.number, a_new.number ) + 2 +
	    	   grid_distance( b.number, b_new.number ) + 1 +
		   grid_distance( b.number, b_new.number ) + 2;
	    if( dis > dis1 ) dis = dis1;

	    dis2 = grid_distance( b.number, b_new.number ) + 2 +
	    	   grid_distance( b_new.number, a.number ) +
		   grid_distance( a.number, a_new.number ) + 2;
	    if( b_new.number == a.number ) dis2--;
	    if( dis > dis2 ) dis = dis2;

	    dis3 = grid_distance( b.number, b_new.number ) + 2 +
	    	   grid_distance( a.number, a_new.number ) + 1 +
		   grid_distance( a.number, a_new.number ) + 2;
	    if( dis > dis3 ) dis = dis3;

            return( dis );
        }
        else
        if( b.type == HELD && b_new.type == GRID ) {
            dis = grid_distance( a.number, a_new.number ) + 3 +
                  grid_distance( b_new.number, a.number );
            if( b_new.number == a.number ) dis--;
            return( dis );
        }
        else
        if( b.type == GRID && b_new.type == HELD ) {
            dis = grid_distance( a.number, a_new.number ) + 3 +
                  grid_distance( a_new.number, b.number );
            if( a_new.number == b.number ) dis--;
            return( dis );
        }
        else
        if( b.type == HELD && b_new.type == HELD ) {
            dis = 2 * grid_distance( a.number, a_new.number ) + 3 ;
            return( dis );
        }
    }
    else
    if( a.type == HELD && a_new.type == GRID ) {
        if( b.type == GRID && b_new.type == GRID ) {
            if( b.number == b_new.number ) return( 0 );
            dis = grid_distance( b.number, b_new.number ) + 3 +
                  grid_distance( a_new.number, b.number );
            if( a_new.number == b.number ) dis--;
            return( dis );
        }
        else
        if( b.type == GRID && b_new.type == HELD ) {
            dis = grid_distance( a_new.number, b.number );
            if( a_new.number == b.number ) dis += 1;
            else                           dis += 2;
            return( dis );
        }
    }
    else
    if( a.type == GRID && a_new.type == HELD ) {
        if( b.type == GRID && b_new.type == GRID ) {
            if( b.number == b_new.number) return( 0 );
            dis = grid_distance( b.number, b_new.number ) + 3 +
                  grid_distance( b_new.number, a.number );
            if( b_new.number == a.number ) dis--;
            return( dis );
        }
        else
        if( b.type == HELD && b_new.type == GRID ) {
            dis = grid_distance( b_new.number, a.number );
            if( b_new.number == a.number ) dis += 1;
            else                           dis += 2;
            return( dis );
        }
    }
    else
    if( a.type == HELD && a_new.type == HELD ) {
        if( b.type == GRID && b_new.type == GRID ) {
            if( b.number == b_new.number ) return( 0 );
            dis = 2 * grid_distance( b.number, b_new.number ) + 3;
            return( dis );
        }
    }
    return( 0 );
}

/*
 *  Determine a lower bound on the number of time steps that would
 *  be needed to change the values of a key variable and the robot
 *  variable.
 */
static int
robot_distance2( a, a_new, r, r_new )
    VALA        a, a_new, r, r_new;
{
    int         dis;

    if( a.type == GRID && a_new.type == GRID ) {
        if( a.number == a_new.number ) return( 0 );
        dis = key_distance( a, a_new ) +
              grid_distance( r.number, a.number ) +
              grid_distance( a_new.number, r_new.number );
        return( dis );
    }
    else
    if( a.type == HELD && a_new.type == GRID ) {
        dis = grid_distance( r.number, a_new.number ) +
              grid_distance( a_new.number, r_new.number ) + 1;
        return( dis );
    }
    else
    if( a.type == GRID && a_new.type == HELD ) {
        dis = grid_distance( r.number, a.number ) +
              grid_distance( a.number, r_new.number ) + 1;
        return( dis );
    }
    return( 0 );
}

/*
 *  check binary key distance constraint.
 */
static int
check_key_distance_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;

    i = C->scheme[r][0];  /* keyi(t1) */
    j = C->scheme[r][1];  /* keyi(t2), 0 <= t1 < t2 <= time */

    a = solution[i];
    b = solution[j];

    /*
     *  actual distance >= lower bound on distance
     */
    return(
        (C->variable_attribute[j].state - C->variable_attribute[i].state) >=
         key_distance( C->value_attribute[i][a], C->value_attribute[j][b] )
    );
}

/*
 *  check 4-ary key distance constraint.
 */
static int
check_key_distance2_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i0, i1, i2, i3;
    int         a0, a1, a2, a3;

    i0 = C->scheme[r][0];  /* keyi(t1)          */
    i1 = C->scheme[r][1];  /* keyj(t1)          */
    i2 = C->scheme[r][2];  /* keyi(t2) t1 < t2  */
    i3 = C->scheme[r][3];  /* keyj(t2)          */

    a0 = solution[i0];
    a1 = solution[i1];
    a2 = solution[i2];
    a3 = solution[i3];

    /*
     *  actual distance >= lower bound on distance
     */
    return(
        (C->variable_attribute[i2].state - C->variable_attribute[i0].state) >=
         key_distance2( C->value_attribute[i0][a0], C->value_attribute[i2][a2],
                        C->value_attribute[i1][a1], C->value_attribute[i3][a3] )
    );
}

/*
 *  check tenary key distance constraint.
 */
static int
check_key_distance3_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i0, i1, i2;
    int         a0, a1, a2;

    i0 = C->scheme[r][0];  /* keyi(t1)   */
    i1 = C->scheme[r][1];  /* lockj(t1+1)  */
    i2 = C->scheme[r][2];  /* keyi(t2)   */

    a0 = solution[i0];
    a1 = solution[i1];
    a2 = solution[i2];

    /*
     *  if keyi locates in a locked grid at time t1 and the grid 
     *  is still locked at time t1+1, the number of time steps
     *  to change the value of keyi should include the steps to 
     *  open the lock.
     */
    if( C->value_attribute[i1][a1].type == LOCKED &&
        !same_attribute( C->value_attribute[i0][a0],
                         C->value_attribute[i2][a2] ) &&
        C->value_attribute[i0][a0].type == GRID &&
        C->value_attribute[i0][a0].number ==
        C->value_attribute[i1][a1].number &&
        ( ( C->variable_attribute[i2].state -
            C->variable_attribute[i0].state ) <=
          ( key_distance( C->value_attribute[i0][a0],
                          C->value_attribute[i2][a2] ) + 2 ) ) ) {
        return( 0 );
    }
    return( 1 );
}

/*
 *  check binary robot distance constraint.
 */
static int
check_robot_distance_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;

    i = C->scheme[r][0];  /* robot(t) */
    j = C->scheme[r][1];  /* robot(t+1) */

    a = solution[i];
    b = solution[j];

    /*
     *  actual distance >= lower bound on distance
     */
    return( grid_distance( C->value_attribute[i][a].number,
                           C->value_attribute[j][b].number ) <= 1 );
}

/*
 *  check 4-ary robot distance constraint.
 */
static int
check_robot_distance2_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i0, i1, i2, i3;
    int         a0, a1, a2, a3;

    i0 = C->scheme[r][0];  /* keyi(t1)          */
    i1 = C->scheme[r][1];  /* robot(t1)         */
    i2 = C->scheme[r][2];  /* keyi(t2) t1 < t2  */
    i3 = C->scheme[r][3];  /* robot(t2)         */

    a0 = solution[i0];
    a1 = solution[i1];
    a2 = solution[i2];
    a3 = solution[i3];

    /*
     *  actual distance >= lower bound on distance
     */
    return(
        (C->variable_attribute[i2].state - C->variable_attribute[i0].state) >=
         robot_distance2( C->value_attribute[i0][a0], C->value_attribute[i2][a2],
                          C->value_attribute[i1][a1], C->value_attribute[i3][a3] )
    );
}

/*
 *  check 4-ary robot distance constraint.
 */
static int
check_robot_distance3_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i0, i1, i2, i3;
    int         a0, a1, a2, a3;
    int         i, key, found;

    i0 = C->scheme[r][0];  /* robot(t1)     */
    i1 = C->scheme[r][1];  /* arm(t1)       */
    i2 = C->scheme[r][2];  /* robot(t1+1)   */
    i3 = C->scheme[r][3];  /* arm(t1+1)     */

    a0 = solution[i0];
    a1 = solution[i1];
    a2 = solution[i2];
    a3 = solution[i3];

    if( !same_attribute( C->value_attribute[i0][a0],
                         C->value_attribute[i1][a1] ) ||
        !same_attribute( C->value_attribute[i1][a1],
                         C->value_attribute[i3][a3] ) ) {
        return( 1 );
    }
    if( C->value_attribute[i1][a1].type == EMPTY ) {
        return( 0 );
    }

    /*
     *  if the robot stays somewhere and does not pick up or
     *  put down some keys, it must have opened some locks 
     *  and thus it can only stay in the grids neighbor
     *  to a lock having the same shape as the key the robot holds.
     */
    found = 0;
    key = C->value_attribute[i1][a1].number;

    for( i = 1; i <= get_num_locks() && !found; i++ ) {
        if( get_key_shape( key ) == get_lock_shape( i ) ) {
            if( grid_distance( C->value_attribute[i0][a0].number,
                               get_lock_orig( i ) ) == 1 ) {
                found = 1;
            }


        }
    }

    return( found );
}

/*
 *  State constraint says the fact that a key is held by the arm
 *  is equivalent to the arm holds the key.
 */
static int
check_state_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;

    i = C->scheme[r][0];  /* keyi(t) */
    j = C->scheme[r][1];  /* arm(t)  */

    a = solution[i];
    b = solution[j];

    if( ( C->value_attribute[i][a].type == HELD ) !=
        ( C->value_attribute[j][b].type == KEY &&
          C->value_attribute[j][b].number ==
          C->variable_attribute[i].number ) ) {
        return( 0 );
    }

    return( 1 );
}

/*
 *  detent constraint 
 */
static int
check_detent_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;

    i = C->scheme[r][0];  /* keyi(t), where keyi(0) == lockj */
    j = C->scheme[r][1];  /* lockj(t)  */

    a = solution[i];
    b = solution[j];

    /*
     *  keyi cannot be moved until lockj is opened.
     */
    if( C->value_attribute[j][b].type == LOCKED &&
        !( C->value_attribute[i][a].type == GRID &&
           C->value_attribute[i][a].number ==
           C->value_attribute[j][b].number ) ) {
        return( 0 );
    }
    return( 1 );
}

/*
 *  Noentry constraint
 */
static int
check_noentry_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;

    i = C->scheme[r][0];  /* keyi(t) where keyi(0) != lockj */
    j = C->scheme[r][1];  /* lockj(t-1)          */

    a = solution[i];
    b = solution[j];

    /*
     *  keyi or the robot cannot enter lockj until it is unlocked.
     */
    if( C->value_attribute[i][a].type == GRID &&
        C->value_attribute[i][a].number ==
        C->value_attribute[j][b].number &&
        C->value_attribute[j][b].type == LOCKED ) {
        return( 0 );
    }
    return( 1 );
}

/*
 *  Action move.
 */
static int
check_action_move_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i0, i1;
    int         a0, a1;

    i0 = C->scheme[r][0];  /* lockj(t) */
    i1 = C->scheme[r][1];  /* robot(t+1) */

    a0 = solution[i0];
    a1 = solution[i1];

    /*
     *  
     */
    if( C->value_attribute[i1][a1].number ==
        C->value_attribute[i0][a0].number &&
        C->value_attribute[i0][a0].type == LOCKED ) {
        return( 0 );
    }
    return( 1 );
}

/*
 *  Action pickup
 */ 
static int
check_action_pickup_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i0, i1, i2, i3, i4, i5;
    int         a0, a1, a2, a3, a4, a5;

    i0 = C->scheme[r][0];  /* robot(t)  */
    i1 = C->scheme[r][1];  /* keyi(t)   */
    i2 = C->scheme[r][2];  /* arm(t)    */
    i3 = C->scheme[r][3];  /* robot(t+1)  */
    i4 = C->scheme[r][4];  /* keyi(t+1)   */
    i5 = C->scheme[r][5];  /* arm(t+1)    */

    a0 = solution[i0];
    a1 = solution[i1];
    a2 = solution[i2];
    a3 = solution[i3];
    a4 = solution[i4];
    a5 = solution[i5];

    if( C->value_attribute[i2][a2].type == EMPTY &&   /* arm(t)   = empty */
        C->value_attribute[i5][a5].type == KEY &&     /* arm(t+1) = keyi  */
        C->value_attribute[i5][a5].number ==
                C->variable_attribute[i1].number &&
        !( C->value_attribute[i1][a1].type == GRID && 
           C->value_attribute[i4][a4].type == HELD && 
           C->value_attribute[i0][a0].number ==       /* robot(t)   = keyi(t) */
                C->value_attribute[i1][a1].number &&  
           C->value_attribute[i0][a0].number ==       /* robot(t+1) = keyi(t) */
                C->value_attribute[i3][a3].number ) ) {
        return( 0 );
    }
    else
    if( C->value_attribute[i1][a1].type == GRID &&
        C->value_attribute[i4][a4].type == HELD &&
        !( C->value_attribute[i5][a5].type == KEY &&  /* arm(t+1) = keyi  */
           C->value_attribute[i5][a5].number ==
                C->variable_attribute[i1].number &&
           C->value_attribute[i0][a0].number ==       /* robot(t) = keyi(t) */
                C->value_attribute[i1][a1].number &&
           C->value_attribute[i0][a0].number ==       /* robot(t+1) = keyi(t) */
                C->value_attribute[i3][a3].number ) ) {
        return( 0 );
    }
    else {
        return( 1 );
    }
}

/*
 *  Action pickup and loose
 */
static int
check_action_pickup_loose_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i0, i1, i2, i3, i4, i5, i6, i7;
    int         a0, a1, a2, a3, a4, a5, a6, a7;

    i0 = C->scheme[r][0];  /* robot(t)  */
    i1 = C->scheme[r][1];  /* keyi(t)   */
    i2 = C->scheme[r][2];  /* arm(t)    */
    i3 = C->scheme[r][3];  /* keyj(t)   */
    i4 = C->scheme[r][4];  /* robot(t+1)  */
    i5 = C->scheme[r][5];  /* keyi(t+1)   */
    i6 = C->scheme[r][6];  /* arm(t+1)    */
    i7 = C->scheme[r][7];  /* keyj(t+1)   */

    a0 = solution[i0];
    a1 = solution[i1];
    a2 = solution[i2];
    a3 = solution[i3];
    a4 = solution[i4];
    a5 = solution[i5];
    a6 = solution[i6];
    a7 = solution[i7];

    if( C->value_attribute[i2][a2].type == KEY &&  /* arm(t) = keyj */
        C->value_attribute[i2][a2].number ==
        C->variable_attribute[i3].number &&
        C->value_attribute[i6][a6].type == KEY &&  /* arm(t+1) = keyi */
        C->value_attribute[i6][a6].number ==
        C->variable_attribute[i1].number &&
        !( C->value_attribute[i0][a0].number ==  /* robot(t) = robot(t+1) */
           C->value_attribute[i4][a4].number &&
           C->value_attribute[i0][a0].number ==  /* robot(t) = keyi(t) */
           C->value_attribute[i1][a1].number &&
           C->value_attribute[i0][a0].number ==  /* robot(t) = keyj(t+1) */
           C->value_attribute[i7][a7].number &&
           C->value_attribute[i3][a3].type == HELD &&
           C->value_attribute[i7][a7].type == GRID &&
           C->value_attribute[i1][a1].type == GRID &&
           C->value_attribute[i5][a5].type == HELD ) ) {
        return( 0 );
    }
    else
    if( C->value_attribute[i2][a2].type == KEY &&  /* arm(t) = keyi */
        C->value_attribute[i2][a2].number ==
        C->variable_attribute[i1].number &&
        C->value_attribute[i6][a6].type == KEY &&  /* arm(t+1) = keyj */
        C->value_attribute[i6][a6].number ==
        C->variable_attribute[i3].number &&
        !( C->value_attribute[i0][a0].number ==  /* robot(t) = robot(t+1) */
           C->value_attribute[i4][a4].number &&
           C->value_attribute[i0][a0].number ==  /* robot(t) = keyj(t) */
           C->value_attribute[i3][a3].number &&
           C->value_attribute[i0][a0].number ==  /* robot(t) = keyi(t+1) */
           C->value_attribute[i5][a5].number &&
           C->value_attribute[i1][a1].type == HELD &&
           C->value_attribute[i5][a5].type == GRID &&
           C->value_attribute[i3][a3].type == GRID &&
           C->value_attribute[i7][a7].type == HELD ) ) {
        return( 0 );
    }
    return( 1 );
}

/* 
 *  Action putdown
 */
static int
check_action_putdown_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i0, i1, i2, i3, i4, i5;
    int         a0, a1, a2, a3, a4, a5;

    i0 = C->scheme[r][0];  /* robot(t)  */
    i1 = C->scheme[r][1];  /* keyi(t)   */
    i2 = C->scheme[r][2];  /* arm(t)    */
    i3 = C->scheme[r][3];  /* robot(t+1)  */
    i4 = C->scheme[r][4];  /* keyi(t+1)   */
    i5 = C->scheme[r][5];  /* arm(t+1)    */

    a0 = solution[i0];
    a1 = solution[i1];
    a2 = solution[i2];
    a3 = solution[i3];
    a4 = solution[i4];
    a5 = solution[i5];

    if( C->value_attribute[i2][a2].type == KEY &&      /* arm(t)   = keyi */
        C->value_attribute[i5][a5].type == EMPTY &&    /* arm(t+1) = empty */
        C->value_attribute[i2][a2].number ==
                C->variable_attribute[i1].number &&
        !( C->value_attribute[i1][a1].type == HELD &&  
           C->value_attribute[i4][a4].type == GRID &&
           C->value_attribute[i0][a0].number ==        /* robot(t) = keyi(t) */
                C->value_attribute[i4][a4].number &&
           C->value_attribute[i0][a0].number ==        /* robot(t+1) = keyi(t) */
                C->value_attribute[i3][a3].number ) ) {
        return( 0 );
    }
    else
    if( C->value_attribute[i1][a1].type == HELD &&
        C->value_attribute[i4][a4].type == GRID &&
        !( C->value_attribute[i2][a2].type == KEY &&   /* arm(t) = keyi */
           C->value_attribute[i2][a2].number ==
                C->variable_attribute[i1].number &&
           C->value_attribute[i0][a0].number ==        /* robot(t) = keyi(t) */  
                C->value_attribute[i4][a4].number &&                             
           C->value_attribute[i0][a0].number ==        /* robot(t+1) = keyi(t) */
                C->value_attribute[i3][a3].number ) ) {
        return( 0 );
    }
    else {
        return( 1 );
    }
}

/*
 *  Action unlock
 */
static int
check_action_unlock_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i0, i1, i2, i3, i4, i5;
    int         a0, a1, a2, a3, a4, a5;

    i0 = C->scheme[r][0];  /* robot(t)  */
    i1 = C->scheme[r][1];  /* locki(t)   */
    i2 = C->scheme[r][2];  /* arm(t)    */
    i3 = C->scheme[r][3];  /* robot(t+1)  */
    i4 = C->scheme[r][4];  /* locki(t+1)   */
    i5 = C->scheme[r][5];  /* arm(t+1)    */

    a0 = solution[i0];
    a1 = solution[i1];
    a2 = solution[i2];
    a3 = solution[i3];
    a4 = solution[i4];
    a5 = solution[i5];

    if( C->value_attribute[i1][a1].type == LOCKED &&
        C->value_attribute[i4][a4].type == UNLOCKED &&
        !( grid_distance( C->value_attribute[i0][a0].number,  /* is neighbor */
                          C->value_attribute[i1][a1].number ) == 1 &&
           C->value_attribute[i0][a0].number ==      /* robot(t) = robot(t+1) */
                C->value_attribute[i3][a3].number &&
           C->value_attribute[i2][a2].type == KEY && /* arm(t) = arm(t+1) */
           C->value_attribute[i5][a5].type == KEY &&
           C->value_attribute[i2][a2].number ==
                C->value_attribute[i5][a5].number &&
           get_key_shape( C->value_attribute[i2][a2].number ) == /* is shape */
                get_lock_shape( C->variable_attribute[i1].number ) ) ) {
        return( 0 );
    }
    else
    if( C->value_attribute[i1][a1].type == UNLOCKED &&
        C->value_attribute[i4][a4].type == LOCKED ) {
        return( 0 );
    }
    else {
        return( 1 );
    }
}

/*
 *  Useless constraint
 */ 
static int
check_useless_action_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i0, i1;
    int         a0, a1;

    i0 = C->scheme[r][0];  /* locki(t)  */
    i1 = C->scheme[r][1];  /* locki(t+1)   */

    a0 = solution[i0];
    a1 = solution[i1];

    /*
     * Once a lock was opened, it will never be locked again.
     */
    if( C->value_attribute[i0][a0].type == UNLOCKED &&
        C->value_attribute[i1][a1].type == LOCKED ) {
        return( 0 );
    }
    return( 1 );
}
