
/*
 *  CPlan solves planning problems formulated as constraint satisfaction
 *  problems.
 *
 *  Copyright (C) 1999  Peter van Beek and Xinguang Chen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.  See the GNU General Public License
 *  for more details (see the file called Copying or contact the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA).
 */

#include "grid.h"
#include "../Share/global.h"
#include "../Share/csp.h"

/*
 *  Private routines.
 */
static void     set_domain();
static void     set_value_attributes();


/*
 *  Generate a CSP corresponding to a blocks world planning problem.
 */
CSP_type *
generate( time )
    int         time;
{
    CSP_type    *C;
    int         i, j, a, b, t, t1, t2, v;
    int         next_con;

    C = (CSP_type *)new_calloc( 1, sizeof(CSP_type) );

    C->time = time;
    C->n = (time+1) * (get_num_locks() +
                       get_num_keys() +
                       2);

    /*
     *  Allocate space for n variables.
     */
    C->domain_size = new_calloc( C->n+1, sizeof(int) );
    C->variable_attribute = (VARA *)new_calloc( C->n+1, sizeof(VARA) );
    C->value_attribute = (VALA **)new_calloc( C->n+1, sizeof(VALA *) );

    /*
     *  Record the attributes of each variable.
     *  Initialize the domain of each variable.
     */
    for( t = 0; t <= time; t++ ) {
        /*
         *  LOCK variables
         */
        for( i = 1; i <= get_num_locks(); i++ ) {
            v = get_var( C, t, LOCK, i );
            set_variable_attributes( C, v, LOCK, t, i );
            set_domain( C, v );
        }
        /*
         *  KEY variables
         */
        for( i = 1; i <= get_num_keys(); i++ ) {
            v = get_var( C, t, KEY, i );
            set_variable_attributes( C, v, KEY, t, i );
            set_domain( C, v );
        }
        /*
         *  ROBOT variable
         */
        v = get_var( C, t, ROBOT, 1 );
        set_variable_attributes( C, v, ROBOT, t, 1 );
        set_domain( C, v );
        /*
         *  ARM variable
         */
        v = get_var( C, t, ARM, 1 );
        set_variable_attributes( C, v, ARM, t, 1 );
        set_domain( C, v );
    }

    /*
     *  Is the variable type visible? (1) or hidden? (0)
     */
    C->visible[LOCK]  = 0;
    C->visible[KEY]   = 1;
    C->visible[ROBOT] = 1;
    C->visible[ARM]   = 1;
    C->n_visible = (time+1) * (get_num_keys() + 2);

    /*
     * Further tune the domain of the lock variables.
     */
    for( a = 1; a <= get_num_locks(); a++ ) {
        /*
         *  The minumum time steps to open the locki, 
         *  i.e, Min{ robot->keyj + keyj->locki }
         */
        i = HUGE;
        for( b = 1; b <= get_num_keys(); b++ ) {
            if( get_lock_shape( a ) == get_key_shape( b ) ) {
                j = grid_distance( get_lock_orig( a ), get_key_orig( b ) );
                if( j == 0 ) continue;
                j += grid_distance( get_robot_orig(), get_key_orig( b ) );
                if( i > j ) i = j;
            }
        }
        for( t = 1; t <= i && t <= time; t++ ) {
            v = get_var( C, t, LOCK, a );
            free( C->value_attribute[v] );
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
            C->domain_size[v] = 1;
            set_value_attributes( C, v, 0, LOCKED, get_lock_orig( a ) );
        }
    }

    next_con = 0;

    /*
     *  Distance constraint on the same key in different states.
     *  Enforce a lower bound on the number of time steps that would
     *  be needed to change the value of a key variable.
     */
    for( t1 = 0;    t1 <  time; t1++ )
    for( t2 = t1+1; t2 <= time; t2++ ) if ( (t2-t1) < (2*get_size() + 2) ) {
        for( a = 1; a <= get_num_keys(); a++ ) {
            set_constraint( C, next_con, 2, KEY_DISTANCE_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t1, KEY, a );
            C->scheme[next_con][1] = get_var( C, t2, KEY, a );
            next_con++;
        }
    }

    /*
     *  Distance constraint on the robot in different states.
     *  Enforce a lower bound on the number of time steps that would
     *  be needed to change the value of the robot variable.
     */
    for( t1 = 0; t1 < time; t1++ ) {
        set_constraint( C, next_con, 2, ROBOT_DISTANCE_CONSTRAINT );
        C->scheme[next_con][0] = get_var( C, t1,   ROBOT, 1 );
        C->scheme[next_con][1] = get_var( C, t1+1, ROBOT, 1 );
        next_con++;
    }

    /*
     *  Useless constraint on a lock variable in different states.
     *  It says the fact that once a lock is unlocked, it will never
     *  be locked again.
     */
    for( t1 = 0; t1 < time; t1++ ) {
        for( a = 1; a <= get_num_locks(); a++ ) {
            set_constraint( C, next_con, 2, USELESS_ACTION_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t1,   LOCK, a );
            C->scheme[next_con][1] = get_var( C, t1+1, LOCK, a );
            next_con++;
        }
    }

    /*
     *  State constraint on a key variable and the arm variable in each state.
     *  A key held by the arm is equvalent to the arm holding the key.
     */
    for( t1 = 1; t1 <= time; t1++ ) {
        for( a = 1; a <= get_num_keys(); a++ ) {
            set_constraint( C, next_con, 2, STATE_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t1, KEY, a );
            C->scheme[next_con][1] = get_var( C, t1, ARM, 1 );
            next_con++;
        }
    }

    /*
     *  Detent constraint on a key variable a lock variable in each state, where
     *  the key is in the locked position at the start. The key cannot be moved
     *  until the lock is opended
     */
    for( t1 = 1; t1 <= time; t1++ ) {
        for( a = 1; a <= get_num_keys(); a++ ) if( get_key_orig( a ) != get_robot_orig() ) {
            for( b = 1; b <= get_num_locks(); b++ ) {
                if( get_key_orig( a ) == get_lock_orig( b ) ) {
                    set_constraint( C, next_con, 2, DETENT_CONSTRAINT );
                    C->scheme[next_con][0] = get_var( C, t1, KEY,  a );
                    C->scheme[next_con][1] = get_var( C, t1, LOCK, b );
                    next_con++;
                }
            }
        }
    }

    /*
     *  Noentry constraint on a key variable and a lock variable in each
     *  state, where they have different initial position. The key cannot 
     *  move in the lock position until the lock is opened.
     */
    for( t1 = 1; t1 <= time; t1++ ) {
        for( a = 1; a <= get_num_locks(); a++ ) {
            for( b = 1; b <= get_num_keys(); b++ ) {
                if( get_key_orig( b ) != get_lock_orig( a ) ) {
                    set_constraint( C, next_con, 2, NOENTRY_CONSTRAINT );
                    C->scheme[next_con][0] = get_var( C, t1,   KEY,  b );
                    C->scheme[next_con][1] = get_var( C, t1-1, LOCK, a );
                    next_con++;
                }
            }
        }
    }

    /*
     *  Action move constraint on a lock variable and the robot variable
     */
    for( t1 = 0; t1 < time; t1++ ) {
        for( a = 1; a <= get_num_locks(); a++ ) {
            set_constraint( C, next_con, 2, ACTION_MOVE_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t1,   LOCK,  a );
            C->scheme[next_con][1] = get_var( C, t1+1, ROBOT, 1 );
            next_con++;
        }
    }

    /*
     *  Key distance3 constraint on the same key in different state and
     *  a lock variable. It enforces a lower bound on the number of time
     *  steps to change the value of the key variables if the key is in
     *  a locked position at the start.
     */
    for( t1 = 0;    t1 <  time; t1++ )
    for( t2 = t1+1; t2 <= time; t2++ ) {
        for( a = 1; a <= get_num_keys();  a++ )
        for( b = 1; b <= get_num_locks(); b++ ) {
            set_constraint( C, next_con, 3, KEY_DISTANCE3_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t1,   KEY,  a );
            C->scheme[next_con][1] = get_var( C, t1+1, LOCK, b );
            C->scheme[next_con][2] = get_var( C, t2,   KEY,  a );
            next_con++;
        }
    }

    /*
     *  Key distance2 constraint on a pair of key in different states.
     *  Enforce a lower bound on the number of time steps to change 
     *  the value of two key variables.
     */
    for( t1 = 0;    t1 <  time; t1++ )
    for( t2 = t1+1; t2 <= time; t2++ ) {
        for( a = 1;   a <  get_num_keys(); a++ )
        for( b = a+1; b <= get_num_keys(); b++ ) {
            set_constraint( C, next_con, 4, KEY_DISTANCE2_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t1, KEY, a );
            C->scheme[next_con][1] = get_var( C, t1, KEY, b );
            C->scheme[next_con][2] = get_var( C, t2, KEY, a );
            C->scheme[next_con][3] = get_var( C, t2, KEY, b );
            next_con++;
        }
    }

    /*
     *  Key distance2 constraint on a key and the robot in different states.
     *  Enforce a lower bound on the number of time steps to change 
     *  the value of the key variable and the robot variable.
     */
    for( t1 = 0;    t1 <  time; t1++ )
    for( t2 = t1+1; t2 <= time; t2++ ) {
        for( a = 1; a <= get_num_keys(); a++ ) {
            set_constraint( C, next_con, 4, ROBOT_DISTANCE2_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t1, KEY,  a );
            C->scheme[next_con][1] = get_var( C, t1, ROBOT, 1 );
            C->scheme[next_con][2] = get_var( C, t2, KEY,  a );
            C->scheme[next_con][3] = get_var( C, t2, ROBOT, 1 );
            next_con++;
        }
    }

    /*
     *  Robot distance3 on the robot and the arm variable in two successive
     *  steps. Enforces that if the robot and the arm do not changed in 
     *  two steps, some locks must be opened.
     */
    for( t1 = 0; t1 <  time; t1++ ) {
        set_constraint( C, next_con, 4, ROBOT_DISTANCE3_CONSTRAINT );
        C->scheme[next_con][0] = get_var( C, t1,   ROBOT,  1 );
        C->scheme[next_con][1] = get_var( C, t1,   ARM,    1 );
        C->scheme[next_con][2] = get_var( C, t1+1, ROBOT,  1 );
        C->scheme[next_con][3] = get_var( C, t1+1, ARM,    1 );
        next_con++;
    }

    /*
     *  Action pickup constraint on the robot, a key and the arm in
     *  two successive steps.
     */
    for( t1 = 0; t1 < time; t1++ ) {
        for( a = 1; a <= get_num_keys(); a++ ) {
            set_constraint( C, next_con, 6, ACTION_PICKUP_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t1,   ROBOT, 1 );
            C->scheme[next_con][1] = get_var( C, t1,   KEY,   a );
            C->scheme[next_con][2] = get_var( C, t1,   ARM,   1 );
            C->scheme[next_con][3] = get_var( C, t1+1, ROBOT, 1 );
            C->scheme[next_con][4] = get_var( C, t1+1, KEY,   a );
            C->scheme[next_con][5] = get_var( C, t1+1, ARM,   1 );
            next_con++;
        }
    }

    /*
     *  Action putdown constraint on the robot, a key and the arm in
     *  two successive steps.
     */
    for( t1 = 0; t1 < time; t1++ ) {
        for( a = 1; a <= get_num_keys(); a++ ) {
            set_constraint( C, next_con, 6, ACTION_PUTDOWN_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t1,   ROBOT, 1 );
            C->scheme[next_con][1] = get_var( C, t1,   KEY,   a );
            C->scheme[next_con][2] = get_var( C, t1,   ARM,   1 );
            C->scheme[next_con][3] = get_var( C, t1+1, ROBOT, 1 );
            C->scheme[next_con][4] = get_var( C, t1+1, KEY,   a );
            C->scheme[next_con][5] = get_var( C, t1+1, ARM,   1 );
            next_con++;
        }
    }

    /*
     *  Action unlock constraint on the robot, a lock and the arm in
     *  two successive steps.
     */
    for( t1 = 0; t1 < time; t1++ ) {
        for( a = 1; a <= get_num_locks(); a++ ) {
            set_constraint( C, next_con, 6, ACTION_UNLOCK_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t1,   ROBOT,  1 );
            C->scheme[next_con][1] = get_var( C, t1,   LOCK,   a );
            C->scheme[next_con][2] = get_var( C, t1,   ARM,    1 );
            C->scheme[next_con][3] = get_var( C, t1+1, ROBOT,  1 );
            C->scheme[next_con][4] = get_var( C, t1+1, LOCK,   a );
            C->scheme[next_con][5] = get_var( C, t1+1, ARM,    1 );
            next_con++;
        }
    }

    
    /*
     *  Action pickup and loose constraint on the robot, two keys and 
     *  the arm in two successive steps.
     */
    for( t1 = 0; t1 < time; t1++ ) {
        for( a = 1;   a <  get_num_keys(); a++ )
        for( b = a+1; b <= get_num_keys(); b++ ) {
            set_constraint( C, next_con, 8, ACTION_PICKUP_LOOSE_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t1,   ROBOT, 1 );
            C->scheme[next_con][1] = get_var( C, t1,   KEY,   a );
            C->scheme[next_con][2] = get_var( C, t1,   ARM,   1 );
            C->scheme[next_con][3] = get_var( C, t1,   KEY,   b );
            C->scheme[next_con][4] = get_var( C, t1+1, ROBOT, 1 );
            C->scheme[next_con][5] = get_var( C, t1+1, KEY,   a );
            C->scheme[next_con][6] = get_var( C, t1+1, ARM,   1 );
            C->scheme[next_con][7] = get_var( C, t1+1, KEY,   b );
            next_con++;
        }
    }

    C->m = next_con;

    /*
     *  When should the constraint be propagated?
     *
     *  C->checkable[r] = u means that constraint type r is propagated
     *          only when the number of uninstantiated variables in the
     *          scheme of the constraint is less than or equal to u.
     *          Constraint propagation then potentially reduces the
     *          domains of each of the uninstantiated variables.
     *  
     *  In particular,
     *    == 1    means only forward checking done on this constraint
     *    >= 2    means arc consistency checking done on this constraint
     *    == HUGE means constraint is always propagated
     */
    C->checkable[ACTION_MOVE_CONSTRAINT] = HUGE;
    C->checkable[ACTION_PICKUP_CONSTRAINT] = 4; /* was 3 */
    C->checkable[ACTION_PICKUP_LOOSE_CONSTRAINT] = 3;
    C->checkable[ACTION_PUTDOWN_CONSTRAINT] = 4; /* was 3 */
    C->checkable[ACTION_UNLOCK_CONSTRAINT] = 4;
    C->checkable[USELESS_ACTION_CONSTRAINT] = HUGE;
    C->checkable[STATE_CONSTRAINT] = HUGE;
    C->checkable[DETENT_CONSTRAINT] = HUGE;
    C->checkable[NOENTRY_CONSTRAINT] = 1;
    C->checkable[KEY_DISTANCE_CONSTRAINT] = HUGE;
    C->checkable[KEY_DISTANCE2_CONSTRAINT] = 2;
    C->checkable[KEY_DISTANCE3_CONSTRAINT] = 2;
    C->checkable[ROBOT_DISTANCE_CONSTRAINT] = HUGE;
    C->checkable[ROBOT_DISTANCE2_CONSTRAINT] = 2;
    C->checkable[ROBOT_DISTANCE3_CONSTRAINT] = 1;

    construct_index( C );

    return( C );
}

int
get_var( C, state, type, number )
    CSP_type    *C;
    int         state, type, number;
{
    int         i;

    i = state * (get_num_locks() + get_num_keys() + 2);

    if( type == LOCK ) {
        i += number;
    }
    else
    if( type == KEY ) {
        i += get_num_locks() + number;
    }
    else
    if( type == ROBOT ) {
        i += get_num_locks() + get_num_keys() + 1;
    }
    else
    if( type == ARM ) {
        i += get_num_locks() + get_num_keys() + 2;
    }
    return( i );
}

static void
set_value_attributes( C, v, a, type, number )
    CSP_type    *C;
    int         type, number;
{
    C->value_attribute[v][a].type = type;
    C->value_attribute[v][a].number = number;
}

/*
 *  Initialize the domain of a variable.
 *  Record the attributes of each value.
 */
static void
set_domain( C, v )
    CSP_type    *C;
    int         v;
{
    int         state, type, i, j;

    get_variable_attributes( C, v, &type, &state, &i );
    if( type == LOCK ) {
        if( state == 0 ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
            C->domain_size[v] = 1;
            set_value_attributes( C, v, 0, LOCKED, get_lock_orig( i ) );
        }
        else {
            /*
             * Domains for LOCK variables: LOCKED and UNLOCKED
             */
            C->value_attribute[v] = (VALA *)new_calloc( 2, sizeof(VALA) );
            C->domain_size[v] = 2;
            set_value_attributes( C, v, 0, LOCKED, get_lock_orig( i ) );
            set_value_attributes( C, v, 1, UNLOCKED, get_lock_orig( i ) );
        }
    }
    else
    if( type == KEY ) {
        /*
         * KEY variable.
         * Domain is all grids or ROBOT
         */
        if( state == 0 ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
            set_value_attributes( C, v, 0, GRID, get_key_orig( i ) );
            C->domain_size[v] = 1;
        }
        else
        if( state == C->time &&
            get_key_orig( i ) != get_key_dest( i ) ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
            set_value_attributes( C, v, 0, GRID, get_key_dest( i ) );
            C->domain_size[v] = 1;
        }
        else {
            C->value_attribute[v] = (VALA *)new_calloc( get_size()*get_size() + 1,
                                                        sizeof(VALA) );
            C->domain_size[v] = get_size()*get_size() + 1;
            for( j = 0; j < get_size()*get_size(); j++ ) {
                set_value_attributes( C, v, j, GRID, j );
            }
            set_value_attributes( C, v, get_size()*get_size(), HELD, 0 );
        }
    }
    else
    if( type == ROBOT ) {
        /*
         * ROBOT variable.
         * Domain is all grids
         */
        if( state == 0 ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
            C->domain_size[v] = 1;
            set_value_attributes( C, v, 0, GRID, get_robot_orig() );
        }
        else {
            C->value_attribute[v] = (VALA *)new_calloc( get_size()*get_size(), sizeof(VALA) );
            C->domain_size[v] = get_size()*get_size();
            for( j = 0; j < get_size()*get_size(); j++ ) {
                set_value_attributes( C, v, j, GRID, j );
            }
        }
    }
    else
    if( type == ARM ) {
        /*
         * ARM variable.
         * Domain is all keys or EMPTY
         */
        if( state == 0 ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
            C->domain_size[v] = 1;
            set_value_attributes( C, v, 0, EMPTY, 0 );
        }
        else {
            C->value_attribute[v] = (VALA *)new_calloc( get_num_keys()+1, sizeof(VALA) );
            C->domain_size[v] = get_num_keys()+1;
            for( j = 1; j <= get_num_keys(); j++ ) {
                set_value_attributes( C, v, j-1, KEY, j );
            }
            set_value_attributes( C, v, get_num_keys(), EMPTY, 0 );
        }
    }
}

