
/*
 *  CPlan solves planning problems formulated as constraint satisfaction
 *  problems.
 *
 *  Copyright (C) 1999  Peter van Beek and Xinguang Chen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.  See the GNU General Public License
 *  for more details (see the file called Copying or contact the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA).
 */

#include <string.h>
#include <stdlib.h>

#include "grid.h"
#include "../Share/global.h"

#define MAX_STEPS 200


static I_type *L;

static void   reduce_keys();


void
get_instance( problem )
    int         problem;
{
    int         i, j, number, shapes, row, col;
    char        name[50], temp[128], **shape;
    FILE        *fp;

    if( problem < 10 ) sprintf( name, "./Problems/p0%1d", problem );
    else               sprintf( name, "./Problems/p%2d",  problem );

    if( (fp = fopen( name, "r" )) == NULL ) {
        printf( "Could not open file %s\n", name );
        exit( 6 );
    }

    L = (I_type *)new_calloc( 1, sizeof(I_type) );

    /*
     *  Get and set parameters.
     */
    L->problem = problem;
    fscanf( fp, "Parameters:\n" );
    fscanf( fp, "\tname: %s\n", name);
    fscanf( fp, "\tsize: %d\n", &(L->size) );
    fscanf( fp, "\tlocks: %d\n", &(L->locks) );
    fscanf( fp, "\tkeys: %d\n", &(L->keys) );
    fscanf( fp, "\tshapes: %d\n", &(shapes) );

    /*
     *  Get shapes.
     */
    shape = (char **) new_calloc( shapes+1, sizeof(char *) );
    for( i = 1; i <= shapes; i++ ) {
        shape[i] = (char *) new_calloc( 128, sizeof(char) );
    }
    fscanf( fp, "Shapes:\n" );
    for( i = 1; i <= shapes; i++ ) {
        fscanf( fp, "\t%s\n", shape[i] );
    }

    /*
     *  Allocate space to store the initial and goal states.
     */
    L->lock_orig  = new_calloc( L->locks+1, sizeof(int) );
    L->lock_shape = new_calloc( L->locks+1, sizeof(int) );
    L->key_orig   = new_calloc( L->keys+1,  sizeof(int) );
    L->key_shape  = new_calloc( L->keys+1,  sizeof(int) );
    L->key_dest   = new_calloc( L->keys+1,  sizeof(int) );
    L->key_name = new_calloc( L->keys+1, sizeof(int) );

    /*
     *  Get initial state.
     */
    fscanf( fp, "Initial:\n" );
    fscanf( fp, "\t(arm-empty)\n" );

    for( i = 1; i <= L->locks; i++ ) {
        fscanf( fp, "\t(locked node%d-%d)\n", &row, &col );
        fscanf( fp, "\t(lock-shape node%d-%d %s)\n", &row, &col, temp );
        L->lock_orig[i] = row*L->size + col;
        for( j = 1; j <= shapes; j++ ) {
            if( strncmp( shape[j], temp, strlen(shape[j]) ) == 0 ) {
                break;
            }
        }
        if( j > shapes ) {
            printf( "Unknown lock shape %s\n", temp );
            exit( 6 );
        }
        L->lock_shape[i] = j;
    }

    for( i = 0; i < L->size*L->size - L->locks; i++ ) {
        fscanf( fp, "\t(open node%d-%d)\n", &row, &col );
    }

    for( i = 1; i <= L->keys; i++ ) {
        fscanf( fp, "\t(key-shape key%d %s)\n", &number, temp );
        fscanf( fp, "\t(at key%d node%d-%d)\n", &number, &row, &col );
        L->key_orig[i] = row * L->size + col;
	L->key_name[i] = number;
        for( j = 1; j <= shapes; j++ ) {
            if( strncmp( shape[j], temp, strlen(shape[j]) ) == 0 ) {
                break;
            }
        }
        if( j > shapes ) {
            printf( "Unknown key shape %s\n", temp );
            exit( 6 );
        }
        L->key_shape[i] = j;
    }

    fscanf( fp, "\t(at-robot node%d-%d)\n", &row, &col );
    L->robot_orig = row * L->size + col;

    /*
     *  Get goal state.
     */
    fscanf( fp, "Goal:\n" );
    for( i = 1; i <= L->keys; i++ ) {
        L->key_dest[i] = L->key_orig[i];
    }
    while( !feof(fp) ) {
        fscanf( fp, "\t(at key%d node%d-%d)\n", &number, &row, &col );
        L->key_dest[number+1] = row * L->size + col;
    }

    fclose( fp );

    reduce_keys();

    L->grid_table = (int **)new_calloc( L->size*L->size, sizeof(int *) );
    for( i = 0; i < L->size*L->size; i++ ) {
        L->grid_table[i] = new_calloc( L->size*L->size, sizeof(int) );
    }
    for( i = 0; i < L->size*L->size; i++ )
    for( j = 0; j < L->size*L->size; j++ ) {
        L->grid_table[i][j] = abs( i/L->size - j/L->size ) +
                              abs( i%L->size - j%L->size );
    }

    for( i = 1; i <= shapes; i++ ) {
        free( shape[i] );
    }
    free( shape );
}

static void
reduce_keys()
{
    int i, j;
    int         *used, count;

    used = new_calloc( L->keys+1, sizeof( int ) );
    count = 0;
    for( i = 1; i <= L->keys; i++ ) {
        used[i] = 0;
        if( L->key_dest[i] != L->key_orig[i] ) {
            used[i] = 1;
            count++;
            continue;
        }
        for( j = 1; j <= L->locks && !used[i]; j++ ) {
            if( L->key_shape[i] == L->lock_shape[j] ) {
                used[i] = 1;
                count++;
            }
        }
    }

    if( count < L->keys ) {
        i = 0;
        for( j = 1; j <= L->keys; j++ ) {
	    if( used[j] ) {
		i++;
		L->key_orig[i] = L->key_orig[j];
		L->key_dest[i] = L->key_dest[j];
		L->key_shape[i] = L->key_shape[j];
		L->key_name[i] = L->key_name[j];
	    }
        }
        L->keys = count;
    }
}

void
print_instance()
{
    printf( "problem: %2d  (size %d, %d keys)\n",
            L->problem, L->size, L->keys );
}

void
dump_instance()
{
    int         i;

    printf( "problem: %2d  (size %d, %d keys)\n",
            L->problem, L->size, L->keys );

    printf("Initial\n");
    for( i = 1; i <= L->locks; i++ ) {
        printf( "\t(locked node%d-%d)\n", L->lock_orig[i]/L->size,
                        L->lock_orig[i]%L->size );
        printf( "\t(lock-shape node%d-%d %d)\n", L->lock_orig[i]/L->size,
                        L->lock_orig[i]%L->size, L->lock_shape[i] );
    }
    for( i = 1; i <= L->keys; i++ ) {
        printf( "\t(key-shape key%d %d)\n", i-1, L->key_shape[i] );
        printf( "\t(at key%d node%d-%d)\n", i-1, L->key_orig[i]/L->size,
                        L->key_orig[i]%L->size );
    }
    printf("\t(at-robot node%d-%d)\n", L->robot_orig/L->size,
                        L->robot_orig%L->size );

    printf("Goal:\n");
    for( i = 1; i <= L->keys; i++ ) {
        if( L->key_dest[i] != L->key_orig[i]) {
           printf("\t(at key%d node%d-%d)\n", i-1, L->key_dest[i]/L->size,
                        L->key_dest[i]%L->size );
        }
    }
}

int get_size()			{ return( L->size ); }
int get_num_locks()		{ return( L->locks ); }
int get_num_keys()		{ return( L->keys ); }

int get_robot_orig(       )	{ return( L->robot_orig ); }

int get_key_orig(   int i )	{ return( L->key_orig[i] ); }
int get_key_shape(  int i )	{ return( L->key_shape[i] ); }
int get_key_dest(   int i )	{ return( L->key_dest[i] ); }
int get_key_name(   int i )	{ return( L->key_name[i] ); }

int get_lock_orig(  int i )	{ return( L->lock_orig[i] ); }
int get_lock_shape( int i )	{ return( L->lock_shape[i] ); }

/*
 *  Return the distance between two grid locations.
 */
int
grid_distance( loc1, loc2 )
    int		loc1, loc2;
{
    return( L->grid_table[loc1][loc2] );
}

int get_upper_bound()	{ return( MAX_STEPS ); }

int
get_lower_bound()
{
    int         i, a, b, n;
    int         bound;
    int         src[L->keys+1];
    int		dst[L->keys+1];

    src[0] = L->robot_orig;
    n = 1;
    for( i = 1; i <= L->keys; i++ ) {
        if( L->key_orig[i] != L->key_dest[i] ) {
            src[n] = L->key_orig[i];
            dst[n] = L->key_dest[i];
            n++;
        }
    }

    bound = 0;

    if( n == 1) return( 1 );

    a = 2*L->size + 20;
    for( i = 1; i < n; i++ ) {
        b = grid_distance( src[0], src[i] );
        if( a > b ) a = b;
    }
    bound += a;

    for( i = 1; i < n; i++ ) {
        bound += grid_distance( src[i], dst[i] ) + 1;
    }
    return( bound );
}

