
/*
 *  CPlan solves planning problems formulated as constraint satisfaction
 *  problems.
 *
 *  Copyright (C) 1999  Peter van Beek and Xinguang Chen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.  See the GNU General Public License
 *  for more details (see the file called Copying or contact the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA).
 */

#include "gripper.h"
#include "../Share/csp.h"
#include "../Share/global.h"

static int      check_distance_constraint();
static int      check_action_pickup_drop_constraint();
static int      check_symmetry_constraint();
static int      check_useless_action_constraint();
static int	check_useless2_action_constraint();
static int      check_state_constraint();
static int      check_lower_bound_constraint();

/*
 *  Check the given values against the constraint.
 */
int
check_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    checks += C->arity[r];

    switch( C->relation[r] ) {

    case DISTANCE_CONSTRAINT:
        return( check_distance_constraint( C, r, solution ) );

    case ACTION_PICKUP_DROP_CONSTRAINT:
        return( check_action_pickup_drop_constraint( C, r, solution ) );

    case SYMMETRY_CONSTRAINT:
        return( check_symmetry_constraint( C, r, solution ) );

    case USELESS_ACTION_CONSTRAINT:
        return( check_useless_action_constraint( C, r, solution ) );

    case USELESS2_ACTION_CONSTRAINT:
        return( check_useless2_action_constraint( C, r, solution ) );

    case STATE_CONSTRAINT:
        return( check_state_constraint( C, r, solution ) );

    case LOWER_BOUND_CONSTRAINT:
        return( check_lower_bound_constraint( C, r, solution ) );

    default:
        printf("Error in check_constraint.\n");
        exit(6);
    }
}

static int
check_action_pickup_drop_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i0, i1, i2, i3, i4, i5;
    int         a0, a1, a2, a3, a4, a5;

    i0 = C->scheme[r][0]; /* ball(t)    */
    i1 = C->scheme[r][1]; /* gripper(t) */
    i2 = C->scheme[r][2]; /* robot(t)   */
    i3 = C->scheme[r][3]; /* ball(t+1)  */
    i4 = C->scheme[r][4]; /* gripper(t+1) */
    i5 = C->scheme[r][5]; /* robot(t+1) */

    a0 = solution[i0];
    a1 = solution[i1];
    a2 = solution[i2];
    a3 = solution[i3];
    a4 = solution[i4];
    a5 = solution[i5];

    if( C->value_attribute[i0][a0].type == ROOM &&
        C->value_attribute[i3][a3].type == GRIPPER &&
        C->value_attribute[i3][a3].number ==
        C->variable_attribute[i1].number &&
        !(C->value_attribute[i1][a1].type == EMPTY &&
          C->value_attribute[i4][a4].type == BALL &&
          C->value_attribute[i4][a4].number ==
          C->variable_attribute[i0].number &&
          C->value_attribute[i2][a2].number ==
          C->value_attribute[i0][a0].number &&
          C->value_attribute[i5][a5].number ==
          C->value_attribute[i0][a0].number) ) {
        return( 0 );
    }
    else
    if( C->value_attribute[i3][a3].type == ROOM &&
        C->value_attribute[i0][a0].type == GRIPPER &&
        C->value_attribute[i0][a0].number ==
        C->variable_attribute[i1].number &&
        !(C->value_attribute[i4][a4].type == EMPTY &&
          C->value_attribute[i1][a1].type == BALL &&
          C->value_attribute[i1][a1].number ==
          C->variable_attribute[i0].number &&
          C->value_attribute[i2][a2].number ==
          C->value_attribute[i3][a3].number &&
          C->value_attribute[i5][a5].number ==
          C->value_attribute[i3][a3].number) ) {
        return( 0 );
    }

    return( 1 );
}

/*
 *  Determine a lower bound on the number of time steps that would
 *  be needed to change the value of a block variable.
 */
static int
distance( a, a_new )
    VALA         a, a_new;
{
    if( a.type == ROOM && a_new.type == ROOM ) {
        if( a.number == a_new.number ) return( 0 );
        else                           return( 3 );
    }
    else if( a.type == ROOM && a_new.type == GRIPPER ) return( 1 );
    else if( a.type == GRIPPER && a_new.type == ROOM  ) return( 1 );
    else
    if( a.type == GRIPPER && a_new.type == GRIPPER ) {
        if( a.number == a_new.number)   return( 0 );
        else                            return( 2 );
    }
    else {
        printf("Error in distance_above.\n");
        exit(6);
    }
}

static int
check_distance_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;

    i = C->scheme[r][0];  /* ball(t1) */
    j = C->scheme[r][1];  /* ball(t2), 0 <= t1 < t2 <= time */

    a = solution[i];
    b = solution[j];

    /*
     *  actual distance >= lower bound on distance
     */
    return(
        (C->variable_attribute[j].state - C->variable_attribute[i].state) >=
        distance( C->value_attribute[i][a], C->value_attribute[j][b] )
    );
}

/*
 * A ball in a gripper is equivalent to the gripper holds the ball
 */
static int
check_state_constraint( C, r, solution)
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;

    i = C->scheme[r][0];  /* ball(t1) */
    j = C->scheme[r][1];  /* gripper(t1) */

    a = solution[i];
    b = solution[j];

    if( C->value_attribute[i][a].type == GRIPPER &&
        C->value_attribute[j][b].type == BALL &&
        C->value_attribute[i][a].number ==
        C->variable_attribute[j].number &&
        !(C->value_attribute[j][b].number ==
          C->variable_attribute[i].number ) ) {
        return( 0 );
    }

    return( 1 );
}

static int
check_symmetry_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i0, i1, i2, i3;
    int         a0, a1, a2, a3;
    int		dest_room;

    i0 = C->scheme[r][0]; /* ball a (t)    */
    i1 = C->scheme[r][1]; /* ball b (t) ( b > a)  */
    i2 = C->scheme[r][2]; /* ball a (t+1)  */
    i3 = C->scheme[r][3]; /* ball b (t+1)  */

    a0 = solution[i0];
    a1 = solution[i1];
    a2 = solution[i2];
    a3 = solution[i3];

    /*
     * If ball "a" needs to be moved, it must move first,
     * and it will be gripped by the smaller numbered 
     * gripper.
     */
    dest_room = get_dest_room( C->variable_attribute[i0].number );

    if( C->value_attribute[i0][a0].number == dest_room ) {
	return( 1 );
    }

    if( C->value_attribute[i0][a0].type == ROOM &&
        C->value_attribute[i1][a1].type == ROOM &&
        C->value_attribute[i0][a0].number ==
        C->value_attribute[i1][a1].number &&
        C->value_attribute[i3][a3].type == GRIPPER ) {

        if( C->value_attribute[i2][a2].type == ROOM ) {
            return( 0 );
        }
        else
        if( C->value_attribute[i2][a2].type == GRIPPER &&
            C->value_attribute[i2][a2].number >=
            C->value_attribute[i3][a3].number ) {
            return( 0 );
        }
    }

    return( 1 );
}


static int
check_useless_action_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i0, i1;
    int         a0, a1;
    int		orig_room, dest_room;

    i0 = C->scheme[r][0]; /* ball(t1)          */
    i1 = C->scheme[r][1]; /* ball(t2), t1 < t2 */

    a0 = solution[i0];
    a1 = solution[i1];

    orig_room = get_orig_room( C->variable_attribute[i0].number );
    dest_room = get_dest_room( C->variable_attribute[i0].number );

    /*
     *  Once a ball leaves its original room (in another
     *  other room or held by gripper), it will never
     *  return to that room.
     */
    if( !(C->value_attribute[i0][a0].type == ROOM &&
          C->value_attribute[i0][a0].number == orig_room ) &&
        C->value_attribute[i1][a1].type == ROOM &&
        C->value_attribute[i1][a1].number == orig_room ) {
        return( 0 );
    }

    /*
     *  Once a ball reaches its destination, it will
     *  never leave that room.
     */
    if( !(C->value_attribute[i1][a1].type == ROOM &&
          C->value_attribute[i1][a1].number == dest_room ) &&
        C->value_attribute[i0][a0].type == ROOM &&
        C->value_attribute[i0][a0].number == dest_room ) {
        return( 0 );
    }

    return( 1 );
}

/*
 * Once a ball is picked up by a gripper, the gripper must put
 * it down in 2 steps.
 */
static int
check_useless2_action_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i0, i1, i2;
    int         a0, a1, a2;

    i0 = C->scheme[r][0]; /* ball a (t)    */
    i1 = C->scheme[r][1]; /* ball a (t+1)  */
    i2 = C->scheme[r][2]; /* ball a (t+2)  */

    a0 = solution[i0];
    a1 = solution[i1];
    a2 = solution[i2];

    if( C->value_attribute[i0][a0].type == GRIPPER &&
	C->value_attribute[i1][a1].type == GRIPPER &&
	C->value_attribute[i2][a2].type == GRIPPER ) {
	return( 0 );
    }

    return( 1 );
}

/*
 * lower bound on number of time steps from current state of balls.
 */
static int
check_lower_bound_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, dest_room;
    int         state, var, val;
    int         bound1, bound2;
    int         moves[3], grippers;

    /*
     * Calculate the number of balls in each room that
     * need to be moved.
     */
    moves[1] = moves[2] = 0;
    for( i = 0; i < C->arity[r]; i++) {
        var = C->scheme[r][i];
        val = solution[var];

	dest_room = get_dest_room( C->variable_attribute[var].number );

        if( C->value_attribute[var][val].type == ROOM &&
            C->value_attribute[var][val].number != dest_room ) {
            moves[C->value_attribute[var][val].number] ++;
        }
    }

    state = C->variable_attribute[C->scheme[r][0]].state;
    grippers = get_num_grippers();

/*
    the lower bound of steps to move those balls to their destinations.
*/

    if( ( moves[1] % grippers ) > 0 ) {
        bound1 = moves[1] / grippers + 1;
    }
    else {
        bound1 = moves[1] / grippers;
    }
    if( ( moves[2] % grippers ) > 0 ) {
        bound2 = moves[2] / grippers + 1;
    }
    else {
        bound2 = moves[2] / grippers;
    }

    bound1 = bound1 * 4 - 1;
    bound2 = bound2 * 4 - 1;
/*
 *  actual distance >= lower bound
 */

    return( ( bound1 <= ( C->time - state ) ) &&
            ( bound2 <= ( C->time - state ) ) );
}
