
/*
 *  CPlan solves planning problems formulated as constraint satisfaction
 *  problems.
 *
 *  Copyright (C) 1999  Peter van Beek and Xinguang Chen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.  See the GNU General Public License
 *  for more details (see the file called Copying or contact the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA).
 */

#include "gripper.h"
#include "../Share/csp.h"
#include "../Share/global.h"

/*
 *  Private routines.
 */
static void     set_domain();
static void     set_value_attributes();


/*
 *  Generate a CSP corresponding to a blocks world planning problem.
 */
CSP_type *
generate( time )
    int         time;
{
    CSP_type    *C;
    int         i, a, b, t, t1, t2, v;
    int         next_con;

    C = (CSP_type *)new_calloc( 1, sizeof(CSP_type) );

    C->time = time;
    C->n = (time+1) * (get_num_balls() +
		       get_num_grippers() +
		       1);

    /*
     *  Allocate space for n variables.
     */
    C->domain_size = new_calloc( C->n+1, sizeof(int) );
    C->variable_attribute = (VARA *)new_calloc( C->n+1, sizeof(VARA) );
    C->value_attribute = (VALA **)new_calloc( C->n+1, sizeof(VALA *) );

    /*
     *  Record the attributes of each variable.
     *  Initialize the domain of each variable.
     */
    for( t = 0; t <= time; t++ ) {
        /*
         *  Ball variables
         */
        for( i = 1; i <= get_num_balls(); i++ ) {
            v = get_var( C, t, BALL, i );
            set_variable_attributes( C, v, BALL, t, i );
            set_domain( C, v );
        }
        /*
         *  Gripper variables
         */
        for( i = 1; i <= get_num_grippers(); i++ ) {
            v = get_var( C, t, GRIPPER, i );
            set_variable_attributes( C, v, GRIPPER, t, i );
            set_domain( C, v );
        }
        /*
         *  ROBOT variable
         */
        v = get_var( C, t, ROBOT, 1 );
        set_variable_attributes( C, v, ROBOT, t, 1 );
        set_domain( C, v );
    }

    /*
     *  Is the variable type visible? (1) or hidden? (0)
     */
    C->visible[BALL]    = 1;
    C->visible[GRIPPER] = 0;
    C->visible[ROBOT]   = 0;
    C->n_visible = (time+1) * get_num_balls();


    next_con = 0;

    /*
     *  Distance constraint on the same ball in different states.
     *  Enforce a lower bound on the number of time steps that would
     *  be needed to change the value of a ball variable.
     */
    for( t1 = 0;    t1 <  time; t1++ )
    for( t2 = t1+1; t2 <= time; t2++ ) if ( t2-t1 < 4) {
        for( a = 1; a <= get_num_balls(); a++ ) {
            set_constraint( C, next_con, 2, DISTANCE_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t1, BALL, a );
            C->scheme[next_con][1] = get_var( C, t2, BALL, a );
            next_con++;
        }
    }

    /*
     *  Uselsss constraint on the same ball in different states.
     *  Prevent some useless actions in the planning, such as
     *  a ball will never return its original room once it leaves the
     *  room.
     */
    for( t1 = 0;    t1 <  time; t1++ )
    for( t2 = t1+1; t2 <= time; t2++ ) {
        for( a = 1; a <= get_num_balls(); a++ ) {
            set_constraint( C, next_con, 2, USELESS_ACTION_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t1, BALL, a );
            C->scheme[next_con][1] = get_var( C, t2, BALL, a );
            next_con++;
        }
    }

    /*
     *  State constraint on a ball variable and a gripper varialbe in
     *  each state. It states the fact that: once a ball is held by a 
     *  gripper, the gripper must take the ball at the same time.
     */
    for( t1 = 1;    t1 <= time - 1; t1++ ) {
        for( a = 1; a <= get_num_balls();    a++ )
        for( b = 1; b <= get_num_grippers(); b++ ) {
            set_constraint( C, next_con, 2, STATE_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t1, BALL,    a );
            C->scheme[next_con][1] = get_var( C, t1, GRIPPER, b );
            next_con++;
        }
    }

    /*
     *  Useless2 constraint on a ball variable over 3 successive states.
     *  It checks that once a ball is held by a gripper, it must
     *  be put down in the next 2 steps.
     */

    for( t1 = 1; t1 <= time - 3; t1++ ) {
        for( a = 1; a <= get_num_balls(); a++ ) {
            set_constraint( C, next_con, 3, USELESS2_ACTION_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t1,   BALL, a );
            C->scheme[next_con][1] = get_var( C, t1+1, BALL, a );
	    C->scheme[next_con][2] = get_var( C, t1+2, BALL, a );
            next_con++;
        }
    }


    /*
     *  Symmetry constraint acts over 2 ball variables at each state.
     *  It enforces that the ball with smaller number must be moved 
     *  before the ball of the larger number. In case that they are
     *  both being picked up by grippers. The "smaller" ball must
     *  be picked up by a "smaller" gripper.
     */
    for( t1 = 0;  t1 <  time; t1++ ) {
        for( a = 1;   a <= get_num_balls(); a++ )
        for( b = a+1; b <= get_num_balls(); b++ ) {
            set_constraint( C, next_con, 4, SYMMETRY_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t1,   BALL, a );
            C->scheme[next_con][1] = get_var( C, t1,   BALL, b );
            C->scheme[next_con][2] = get_var( C, t1+1, BALL, a );
            C->scheme[next_con][3] = get_var( C, t1+1, BALL, b );
            next_con++;
        }
    }

    /*
     *  Pickup and Drop Actions on the balls, grippers and the robot 
     *  at each state.
     */
    for( t = 0; t < time; t++ ) {
        for( a = 1; a <= get_num_balls(); a++ )
        for( b = 1; b <= get_num_grippers(); b++ ) {
            set_constraint( C, next_con, 6, ACTION_PICKUP_DROP_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t,   BALL,    a );
            C->scheme[next_con][1] = get_var( C, t,   GRIPPER, b );
            C->scheme[next_con][2] = get_var( C, t,   ROBOT,   1 );
            C->scheme[next_con][3] = get_var( C, t+1, BALL,    a );
            C->scheme[next_con][4] = get_var( C, t+1, GRIPPER, b );
            C->scheme[next_con][5] = get_var( C, t+1, ROBOT,   1 );
            next_con++;
        }
    }

    /*
     *  Lower bound constraint over all ball variables at each state.
     *  It enforces a lower bound on number of time steps to move the 
     *  balls to their destinations.
     */
    for( t = 0; t < time; t++ ) {
        set_constraint( C, next_con, get_num_balls(), LOWER_BOUND_CONSTRAINT );
        for( a = 1; a <= get_num_balls(); a++ ) {
            C->scheme[next_con][a-1] = get_var( C, t, BALL, a );
        }
        next_con++;
    }


    C->m = next_con;

    /*
     *  When should the constraint be propagated?
     *
     *  C->checkable[r] = u means that constraint type r is propagated
     *          only when the number of uninstantiated variables in the
     *          scheme of the constraint is less than or equal to u.
     *          Constraint propagation then potentially reduces the
     *          domains of each of the uninstantiated variables.
     *  
     *  In particular,
     *    == 1    means only forward checking done on this constraint
     *    >= 2    means arc consistency checking done on this constraint
     *    == HUGE means constraint is always propagated
     */
    C->checkable[DISTANCE_CONSTRAINT] = 2;		/* arity = 2 */
    C->checkable[STATE_CONSTRAINT] = 2;			/* arity = 2 */
    C->checkable[USELESS_ACTION_CONSTRAINT] = 2;	/* arity = 2 */
    C->checkable[USELESS2_ACTION_CONSTRAINT] = 2;	/* arity = 3 */
    C->checkable[SYMMETRY_CONSTRAINT] = 3;		/* arity = 4 */
    C->checkable[ACTION_PICKUP_DROP_CONSTRAINT] = 5;	/* arity = 6 */
    C->checkable[LOWER_BOUND_CONSTRAINT] = 1;		/* arity = varying */

    construct_index( C );

    return( C );
}

int
get_var( C, state, type, number )
    CSP_type    *C;
    int         state, type, number;
{
    int         i;

    i = state * (get_num_balls() + get_num_grippers() + 1);

    if( type == BALL ) {
        i += number;
    }
    else
    if( type == GRIPPER ) {
        i += get_num_balls() + number;
    }
    else
    if( type == ROBOT ) {
        i += get_num_balls() + get_num_grippers() + number;
    }
    return( i );
}

static void
set_value_attributes( C, v, a, type, number )
    CSP_type    *C;
    int         type, number;
{
    C->value_attribute[v][a].type = type;
    C->value_attribute[v][a].number = number;
}

/*
 *  Initialize the domain of a variable.
 *  Record the attributes of each value.
 */
static void
set_domain( C, v )
    CSP_type    *C;
    int         v;
{
    int         state, type, i, j;

    get_variable_attributes( C, v, &type, &state, &i );
    if( type == BALL ) {
        if( state == 0 ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
            C->domain_size[v] = 1;
            set_value_attributes( C, v, 0, ROOM, get_orig_room( i ) );
        }
        else
        if( state == C->time ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
            C->domain_size[v] = 1;
            set_value_attributes( C, v, 0, ROOM, get_dest_room( i ) );
        }
        else {
            /*
             * Domains for BALL variables: 0, 1, 2 or 3, rooms or grippers.
             */
            C->value_attribute[v] = (VALA *)new_calloc( 2 + get_num_grippers(),
                                                        sizeof(VALA) );
            C->domain_size[v] = 2 + get_num_grippers();
            set_value_attributes( C, v, 0, ROOM, 1);
            set_value_attributes( C, v, 1, ROOM, 2);
            for( j = 1; j <= get_num_grippers(); j++)
                set_value_attributes( C, v, j+1, GRIPPER, j);
        }
    }
    else
    if( type == GRIPPER ) {
        /*
         * GRIPPER variable.
         * Domain is all balls or EMPTY.
         */
        if( state == 0 ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
            set_value_attributes( C, v, 0, EMPTY, 0 );
            C->domain_size[v] = 1;
        }
        else
        if( state == C->time ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
            set_value_attributes( C, v, 0, EMPTY, 0 );
            C->domain_size[v] = 1;
        }
        else {
            C->value_attribute[v] = (VALA *)new_calloc( get_num_balls() + 1,
                                                        sizeof(VALA) );
            C->domain_size[v] = get_num_balls() + 1;
            for( j = 1; j <= get_num_balls(); j++ )
                set_value_attributes( C, v, j-1, BALL, j );
            set_value_attributes( C, v, get_num_balls(), EMPTY, 0);
        }
    }
    else
    if( type == ROBOT ) {
        /*
         * GRIPPER variable.
         * Domain is all balls or EMPTY.
         */
        if( state == 0 ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
            C->domain_size[v] = 1;
            set_value_attributes( C, v, 0, ROOM, get_robby_orig_room() );
        }
        else {
            C->value_attribute[v] = (VALA *)new_calloc( 2, sizeof(VALA) );
            C->domain_size[v] = 2;
            set_value_attributes( C, v, 0, ROOM, 1);
            set_value_attributes( C, v, 1, ROOM, 2);
        }
    }
}

