
/*
 *  CPlan solves planning problems formulated as constraint satisfaction
 *  problems.
 *
 *  Copyright (C) 1999  Peter van Beek and Xinguang Chen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.  See the GNU General Public License
 *  for more details (see the file called Copying or contact the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA).
 */

#include "logistics.h"
#include "../Share/global.h"
#include "../Share/csp.h"


static int	check_distance_constraint();
static int	check_distance2_constraint();
static int	check_successful_action_constraint();
static int	check_useless_action_constraint();
static int	check_force_parallelism_constraint();
static int	check_force_move_constraint();
static int	check_outgoing_constraint();
static int	check_incoming_constraint();
static int	check_enroute_constraint();
static int	check_upper_bound_constraint();
static int	check_symmetry_constraint();
static int	check_symmetry2_constraint();
static int	check_symmetry3_constraint();
static int	check_symmetry4_constraint();


/*
 *  Check the given values against the constraint.
 */
int
check_constraint( C, rel_index, solution )
    CSP_type    *C;
    int         rel_index;
    int         *solution;
{
    checks += C->arity[rel_index];

    switch( C->relation[rel_index] ) {

    case DISTANCE_CONSTRAINT:
	return( check_distance_constraint( C, rel_index, solution ) );

    case DISTANCE2_CONSTRAINT:
	return( check_distance2_constraint( C, rel_index, solution ) );

    case SUCCESSFUL_LOAD_UNLOAD_CONSTRAINT:
	return( check_successful_action_constraint( C, rel_index, solution ) );

    case USELESS_ACTIONS_CONSTRAINT:
	return( check_useless_action_constraint( C, rel_index, solution ) );

    case FORCE_PARALLELISM_CONSTRAINT:
	return( check_force_parallelism_constraint( C, rel_index, solution ) );

    case OUTGOING_CONSTRAINT:
	return( check_outgoing_constraint( C, rel_index, solution ) );

    case INCOMING_CONSTRAINT:
	return( check_incoming_constraint( C, rel_index, solution ) );

    case ENROUTE_CONSTRAINT:
	return( check_enroute_constraint( C, rel_index, solution ) );

    case UPPER_BOUND_CONSTRAINT:
	return( check_upper_bound_constraint( C, rel_index, solution ) );

    case SYMMETRY_CONSTRAINT:
	return( check_symmetry_constraint( C, rel_index, solution ) );

    case SYMMETRY2_CONSTRAINT:
	return( check_symmetry2_constraint( C, rel_index, solution ) );

    case SYMMETRY3_CONSTRAINT:
	return( check_symmetry3_constraint( C, rel_index, solution ) );

    case SYMMETRY4_CONSTRAINT:
	return( check_symmetry4_constraint( C, rel_index, solution ) );

    case FORCE_MOVE_CONSTRAINT:
	return( check_force_move_constraint( C, rel_index, solution ) );

    default:
	printf("Error in check_constraint.\n");
	exit(6);
    }
}

static int
same_location( a, b )
    VALA	a, b;
{
    return( a.type == LOCATION &&
	    b.type == LOCATION &&
	    a.city == b.city &&
	    a.number == b.number );
}

static int
same_attributes( a, b )
    VALA	a, b;
{
    return( a.type   == b.type &&
	    a.city   == b.city &&
	    a.number == b.number );
}

static int
same_city( a, b )
    VALA	a, b;
{
    return( a.city == b.city );
}

static int
airport_location( a )
    VALA	a;
{
    return( a.type   == LOCATION &&
	    a.number == get_airport_loca( a.city ) );
}

/*
 *  Determine a lower bound on the number of time steps that would
 *  be needed to change the value of a package variable.
 */
int
distance( a, b )
    VALA	 a, b;
{
    int total;

    if( same_attributes( a, b ) ) {
	return( 0 );
    }
    else
    if( a.type == LOCATION && b.type == LOCATION ) {
	if( same_city( a, b ) )
	    return( 3 );
	else {
	    total = 3;
	    if( !airport_location( a ) ) total += 3;
	    if( !airport_location( b ) ) total += 3;
	    return( total );
	}
    }
    else
    if( a.type == LOCATION && b.type == PLANE ) {
	if( airport_location( a ) ) return( 1 );
	else			    return( 4 );
    }
    else
    if( a.type == PLANE && b.type == LOCATION ) {
	if( airport_location( b ) ) return( 1 );
	else			    return( 4 );
    }
    else
    if( a.type == LOCATION && b.type == TRUCK ) {
	if(      same_city( a, b )     ) return( 1 );
	else if( airport_location( a ) ) return( 4 );
	else				 return( 7 );
    }
    else
    if( a.type == TRUCK && b.type == LOCATION ) {
	if(      same_city( a, b )     ) return( 1 );
	else if( airport_location( b ) ) return( 4 );
	else				 return( 7 );
    }
    else
    if( a.type == TRUCK && b.type == TRUCK ) {
	if( same_city( a, b ) ) return( 2 );
	else			return( 5 );
    }
    else {
        /*
         * a.type == PLANE && b.type == PLANE
         * a.type == PLANE && b.type == TRUCK
         * a.type == TRUCK && b.type == PLANE
         */
	return( 2 );
    }
}

/*
 *  Check 2-ary distance constraint.
 */
static int
check_distance_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;

    i = C->scheme[r][0];  /* package(t1) */
    j = C->scheme[r][1];  /* package(t2), 0 <= t1 < t2 <= time */

    a = solution[i];
    b = solution[j];

    /*
     *  actual distance >= lower bound on distance
     */
    return(
	(C->variable_attribute[j].state - C->variable_attribute[i].state) >=
	distance( C->value_attribute[i][a], C->value_attribute[j][b] )
    );

}

/*
 *  Determine a lower bound on the number of time steps that would
 *  be needed to change the value of two package variables.
 */
static int
distance2( a, a_new,
	   b, b_new )
    VALA	a, a_new,
	        b, b_new;
{
    int		airport;

    if( a.type != PLANE &&
        b.type != PLANE &&
        a_new.type == PLANE &&
        b_new.type == PLANE &&
	a_new.number == b_new.number ) {

	if( a.type == LOCATION && b.type == LOCATION ) {

	    airport = 0;
	    if( airport_location( a ) ) airport++;
	    if( airport_location( b ) ) airport++;

	    if( !same_city( a, b ) ) {
		if     ( airport == 0 ) return( 6 );
		else if( airport == 1 ) return( 4 );
		else			return( 3 );
	    }
	    else {
		if     ( airport == 0 && a.number != b.number ) return( 6 );
		else if( airport == 0 && a.number == b.number ) return( 4 );
		else if( airport == 1 ) return( 4 );
		else			return( 1 );
	    }
	}
	else
	if( (a.type == LOCATION && b.type == TRUCK) ||
	    (a.type == TRUCK && b.type == LOCATION) ) {

	    if( !airport_location( a ) &&
		!airport_location( b ) )   return( 4 );
	    else
	    if( !same_city( a, b ) )	   return( 3 );
	    else			   return( 2 );

	}
	else
	if( a.type == TRUCK && b.type == TRUCK ) {

	    if( !same_city( a, b ) ) return( 4 );
	    else		     return( 2 );
	}
    }
    else
    if( a.type == PLANE &&
        b.type == PLANE &&
	a.number == b.number &&
        a_new.type != PLANE &&
        b_new.type != PLANE ) {

	return( distance2( a_new, a,
			   b_new, b ) );
    }

    return( 0 );
}

/*
 *  Check 4-ary distance constraint.
 */
static int
check_distance2_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j, k, l;
    int         a, b, c, d;

    i = C->scheme[r][0];  /* package_a(t1) */
    j = C->scheme[r][1];  /* package_a(t2) */
    k = C->scheme[r][2];  /* package_b(t1) */
    l = C->scheme[r][3];  /* package_b(t2) */

    a = solution[i];
    b = solution[j];
    c = solution[k];
    d = solution[l];

    /*
     *  actual distance >= lower bound on distance
     */
    return(
	(C->variable_attribute[j].state - C->variable_attribute[i].state) >=
	distance2( C->value_attribute[i][a], C->value_attribute[j][b],
		   C->value_attribute[k][c], C->value_attribute[l][d] )
    );
}

/*
 *  Check a 2-ary grouping constraint.
 */
static int
check_outgoing_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;
    int		orig_city;

    i = C->scheme[r][0];  /* package1(t) */
    j = C->scheme[r][1];  /* package2(t) */

    a = solution[i];
    b = solution[j];

    /* get city of origin */
    orig_city = get_orig_city( C->variable_attribute[i].number );

    /*
     *  Two packages should travel together within the
     *  origination city.
     */
    if( ((C->value_attribute[i][a].type == TRUCK &&
	  C->value_attribute[i][a].city == orig_city) ||
         (C->value_attribute[j][b].type == TRUCK &&
          C->value_attribute[j][b].city == orig_city)) &&
	 !same_attributes( C->value_attribute[i][a],
			   C->value_attribute[j][b] ) ) {
        return( 0 );
    }
    else {
        return( 1 );
    }
}

/*
 *  Check a 2-ary grouping constraint.
 */
static int
check_incoming_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;
    int		dest_city;

    i = C->scheme[r][0];  /* package1(t) */
    j = C->scheme[r][1];  /* package2(t) */

    a = solution[i];
    b = solution[j];

    /* get city of destination */
    dest_city = get_dest_city( C->variable_attribute[i].number );

    /*
     *  Two packages should travel together within the
     *  destination city.
     */
    if( ((C->value_attribute[i][a].type == TRUCK &&
	  C->value_attribute[i][a].city == dest_city) ||
         (C->value_attribute[j][b].type == TRUCK &&
          C->value_attribute[j][b].city == dest_city)) &&
	 !same_attributes( C->value_attribute[i][a],
			   C->value_attribute[j][b] ) ) {
        return( 0 );
    }
    else {
        return( 1 );
    }
}

/*
 *  Check a 2-ary grouping constraint.
 */
static int
check_enroute_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;

    i = C->scheme[r][0];  /* package1(t) */
    j = C->scheme[r][1];  /* package2(t) */

    a = solution[i];
    b = solution[j];

    if( (C->value_attribute[i][a].type == PLANE ||
         C->value_attribute[j][b].type == PLANE) &&
	 !same_attributes( C->value_attribute[i][a],
			   C->value_attribute[j][b] ) ) {
        return( 0 );
    }
    else {
        return( 1 );
    }
}

/*
 *  Check a 4-ary constraint.
 */
static int
check_successful_action_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j, k, l, mode;
    int         a, b, c, d;

    i = C->scheme[r][0]; /* package */
    j = C->scheme[r][1]; /* plane or truck */
    k = C->scheme[r][2]; /* package */
    l = C->scheme[r][3]; /* plane or truck */
    mode = C->variable_attribute[j].type;

    a = solution[i];
    b = solution[j];
    c = solution[k];
    d = solution[l];

    /*
     *  For a package to successfully load onto a plane, the plane
     *  and the package must be at the same location and the plane
     *  must still be at that location in the next time step.
     */
    if( C->value_attribute[i][a].type == LOCATION &&
        C->value_attribute[k][c].type == mode &&
        C->value_attribute[k][c].number ==
    	    C->variable_attribute[j].number &&
        !(same_location( C->value_attribute[i][a],
    		         C->value_attribute[j][b] ) &&
          same_location( C->value_attribute[i][a],
    		         C->value_attribute[l][d] )) ) {
	return( 0 );
    }
    /*
     *  For a package to successfully unload from a plane, the plane
     *  and the package must be at the same location in the next time
     *  step and the plane must have been at that location in the
     *  previous time step.
     */
    else
    if( C->value_attribute[i][a].type == mode &&
        C->value_attribute[k][c].type == LOCATION &&
        C->value_attribute[i][a].number ==
    	    C->variable_attribute[j].number &&
        !(same_location( C->value_attribute[k][c],
    		         C->value_attribute[j][b] ) &&
          same_location( C->value_attribute[k][c],
    		         C->value_attribute[l][d] )) ) {
	return( 0 );
    }
    else {
	return( 1 );
    }
}

/*
 *  Check a 3-ary constraint.
 */
static int
check_useless_action_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j, k;
    int         a, b, c;
    int		dest_city;

    i = C->scheme[r][0]; /* package(t1) */
    j = C->scheme[r][1]; /* package(t2) */
    k = C->scheme[r][2]; /* package(t3), 0 <= t1 < t2 < t3 <= time */

    a = solution[i];
    b = solution[j];
    c = solution[k];

    /*
     *  Get the destination city of this package.
     */
    dest_city = get_dest_city( C->variable_attribute[i].number );

    /*
     *  Prevent useless actions constraint.  If a package is at
     *  a {location,plane,truck} at time t it will never return
     *  to that value in subsequent time steps.
     */
    if(  same_attributes( C->value_attribute[i][a],
			  C->value_attribute[k][c] ) &&
	!same_attributes( C->value_attribute[i][a],
			  C->value_attribute[j][b] ) ) {
	return( 0 );
    }
    /*
     *  Once a package leaves a city, it never returns to
     *  any location or truck in that city.
     */
    else
    if( C->value_attribute[i][a].type == LOCATION &&
	C->value_attribute[j][b].type == PLANE &&
	(C->value_attribute[k][c].type == LOCATION ||
	 C->value_attribute[k][c].type == TRUCK) &&
	same_city( C->value_attribute[i][a],
		   C->value_attribute[k][c] ) ) {
	return( 0 );
    }
    /*
     *  Once a package reaches the destination city, it
     *  never leaves it.
     */
    else
    if( C->value_attribute[i][a].type == PLANE &&
	C->value_attribute[j][b].type == LOCATION &&
	C->value_attribute[j][b].city == dest_city &&
	C->value_attribute[k][c].city != dest_city ) {
	return( 0 );
    }
    else {
	return( 1 );
    }
}

/*
 *  Check a 4-ary constraint.
 */
static int
check_force_parallelism_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j, k, l;
    int         a, b, c, d;

    i = C->scheme[r][0]; /* package1(t) */
    j = C->scheme[r][1]; /* package1(t+1) */
    k = C->scheme[r][2]; /* package2(t+1) */
    l = C->scheme[r][3]; /* package2(t+2) */

    a = solution[i];
    b = solution[j];
    c = solution[k];
    d = solution[l];

    /*
     *  Force parallelism constraint.  Two packages cannot be
     *  loaded or unloaded onto the same truck or plane in successive
     *  states; they all must be loaded and unloaded in the same state.
     */
    if( C->value_attribute[i][a].type == LOCATION &&
        C->value_attribute[j][b].type == TRUCK &&
        C->value_attribute[k][c].type == LOCATION &&
        C->value_attribute[l][d].type == TRUCK &&
        C->value_attribute[j][b].number ==
    	C->value_attribute[l][d].number ) {
        /* load, load */
        return( 0 );
    }
    else
    if( C->value_attribute[i][a].type == TRUCK &&
        C->value_attribute[j][b].type == LOCATION &&
        C->value_attribute[k][c].type == TRUCK &&
        C->value_attribute[l][d].type == LOCATION &&
        C->value_attribute[i][a].number ==
    	C->value_attribute[k][c].number ) {
        /* unload, unload */
        return( 0 );
    }
    else
    if( C->value_attribute[i][a].type == LOCATION &&
        C->value_attribute[j][b].type == TRUCK &&
        C->value_attribute[k][c].type == TRUCK &&
        C->value_attribute[l][d].type == LOCATION &&
        C->value_attribute[j][b].number ==
    	C->value_attribute[k][c].number ) {
        /* load, unload */
        return( 0 );
    }
    else
    if( C->value_attribute[i][a].type == TRUCK &&
        C->value_attribute[j][b].type == LOCATION &&
        C->value_attribute[k][c].type == LOCATION &&
        C->value_attribute[l][d].type == TRUCK &&
        C->value_attribute[i][a].number ==
    	C->value_attribute[l][d].number ) {
        /* unload, load */
        return( 0 );
    }
    else
    if( C->value_attribute[i][a].type == LOCATION &&
        C->value_attribute[j][b].type == PLANE &&
        C->value_attribute[k][c].type == LOCATION &&
        C->value_attribute[l][d].type == PLANE &&
        C->value_attribute[j][b].number ==
    	C->value_attribute[l][d].number ) {
        /* load, load */
        return( 0 );
    }
    else
    if( C->value_attribute[i][a].type == PLANE &&
        C->value_attribute[j][b].type == LOCATION &&
        C->value_attribute[k][c].type == PLANE &&
        C->value_attribute[l][d].type == LOCATION &&
        C->value_attribute[i][a].number ==
    	C->value_attribute[k][c].number ) {
        /* unload, unload */
        return( 0 );
    }
    else
    if( C->value_attribute[i][a].type == LOCATION &&
        C->value_attribute[j][b].type == PLANE &&
        C->value_attribute[k][c].type == PLANE &&
        C->value_attribute[l][d].type == LOCATION &&
        C->value_attribute[j][b].number ==
    	C->value_attribute[k][c].number ) {
        /* load, unload */
        return( 0 );
    }
    else
    if( C->value_attribute[i][a].type == PLANE &&
        C->value_attribute[j][b].type == LOCATION &&
        C->value_attribute[k][c].type == LOCATION &&
        C->value_attribute[l][d].type == PLANE &&
        C->value_attribute[i][a].number ==
    	C->value_attribute[l][d].number ) {
        /* unload, load */
        return( 0 );
    }
    else {
        return( 1 );
    }
}


/*
 *  From S_0,    outgoing package, t           >= upper_bound_outgoing( city )
 *  From S_goal, incoming package, (C->time-t) >= upper_bound_incoming( city )
 */
static int
check_upper_bound_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;

    i = C->scheme[r][0]; /* package(initial) or package(goal) */
    j = C->scheme[r][1]; /* outgoing or incoming package(t), 0 < t < time */

    a = solution[i];
    b = solution[j];

    /*
     *  Package has already left the origination city (outgoing) or
     *  package hasn't reached the destination city (incoming)
     *  (it is either in a plane, in a truck in another city,
     *  or at a location in another city.
     */
    if( !same_city( C->value_attribute[i][a],
		    C->value_attribute[j][b] ) ) {
        return( 1 );
    }
    /*
     *  Package is at the airport in the originating city (outgoing) or
     *  package is at the airport in the destination city (incoming).
     */
    else
    if( airport_location( C->value_attribute[j][b] ) &&
	same_city( C->value_attribute[j][b],
		   C->value_attribute[i][a] ) ) {
        return( 1 );
    }
    /*
     *  Package is at a non-airport location or in a truck in the
     *  city of origination (outgoing) or in the city of destination
     *  (incoming), even though the upper bound has been exceeded.
     */
    else {
        return( 0 );
    }
}

/*
 *  Check a 4-ary symmetry constraint.
 */
static int
check_symmetry_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j, k, l;
    int         a, b, c, d;

    i = C->scheme[r][0];  /* package(t), 0 <= t <= time-1 */
    j = C->scheme[r][1];  /* package(t+1)		  */
    k = C->scheme[r][2];  /* package(0)			  */
    l = C->scheme[r][3];  /* plane(0)			  */

    a = solution[i];
    b = solution[j];
    c = solution[k];
    d = solution[l];

    /*
     *  If package is at the airport location in originating city
     *  it must load onto the given plane.
     */
    if( airport_location( C->value_attribute[i][a] ) &&
	same_city( C->value_attribute[i][a],
		   C->value_attribute[k][c] ) &&
        C->value_attribute[j][b].type == PLANE &&
	C->value_attribute[j][b].number !=
		C->variable_attribute[l].number ) {
        return( 0 );
    }
    else {
        return( 1 );
    }

}

/*
 *  Check a 5-ary symmetry constraint.
 */
static int
check_symmetry2_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i0, i1, i2, i3, i4;
    int         a0, a1, a2, a3, a4;

    i0 = C->scheme[r][0]; /* package_a(t1)   */
    i1 = C->scheme[r][1]; /* package_a(t1+1) */
    i2 = C->scheme[r][2]; /* package_b(t1)   */
    i3 = C->scheme[r][3]; /* package_b(t1+1) */
    i4 = C->scheme[r][4]; /* package_b(t2)   */

    a0 = solution[i0];
    a1 = solution[i1];
    a2 = solution[i2];
    a3 = solution[i3];
    a4 = solution[i4];

    /*
     *
     */
    if( same_location( C->value_attribute[i0][a0],
		       C->value_attribute[i2][a2] ) &&
	same_location( C->value_attribute[i0][a0],
		       C->value_attribute[i3][a3] ) &&
        C->value_attribute[i1][a1].type == PLANE &&
        C->value_attribute[i4][a4].type == PLANE &&
	C->value_attribute[i1][a1].number ==
	    C->value_attribute[i4][a4].number ) {

	return( 0 );
    }
    else
    if( same_location( C->value_attribute[i1][a1],
		       C->value_attribute[i2][a2] ) &&
	same_location( C->value_attribute[i1][a1],
		       C->value_attribute[i3][a3] ) &&
        C->value_attribute[i0][a0].type == PLANE &&
        C->value_attribute[i4][a4].type == PLANE &&
	C->value_attribute[i0][a0].number ==
	    C->value_attribute[i4][a4].number ) {

	return( 0 );
    }
    else {
	return( 1 );
    }
}

/*
 *  Check a binary symmetry constraint.
 */
static int
check_symmetry3_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;

    i = C->scheme[r][0]; /* package_a(5) */
    j = C->scheme[r][1]; /* package_b(5) */

    a = solution[i];
    b = solution[j];

    if( C->value_attribute[i][a].type == PLANE &&
        C->value_attribute[j][b].type == PLANE &&
	indistinguishable_planes( C->value_attribute[i][a].number,
				  C->value_attribute[j][b].number ) &&
	!(C->value_attribute[i][a].number <
	    C->value_attribute[j][b].number) ) {
	return( 0 );
    }
    else {
	return( 1 );
    }
}

/*
 *  Check a binary symmetry constraint.
 */
static int
check_symmetry4_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;

    i = C->scheme[r][0]; /* package_a(2) */
    j = C->scheme[r][1]; /* package_b(2) */

    a = solution[i];
    b = solution[j];

    if( C->value_attribute[i][a].type == TRUCK &&
        C->value_attribute[j][b].type == TRUCK &&
	indistinguishable_trucks( C->value_attribute[i][a].number,
				  C->value_attribute[j][b].number ) &&
	!(C->value_attribute[i][a].number <
	    C->value_attribute[j][b].number) ) {
	return( 0 );
    }
    else {
	return( 1 );
    }
}

/*
 *  Check a 4-ary constraint.
 */
static int
check_force_move_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j, k, l;
    int         a, b, c, d;

    i = C->scheme[r][0]; /* package(t)   */
    j = C->scheme[r][1]; /* package(t+1) */
    k = C->scheme[r][2]; /* package(t+2) */
    l = C->scheme[r][3]; /* package(t+3) */

    a = solution[i];
    b = solution[j];
    c = solution[k];
    d = solution[l];

    /*
     *
     *
     */
    if( C->value_attribute[i][a].type == LOCATION &&
        C->value_attribute[j][b].type == PLANE &&
        !(C->value_attribute[k][c].type == PLANE &&
          C->value_attribute[l][d].type == LOCATION) ) {
	return( 0 );
    }
    /*
     *
     *
     */
    else
    if( C->value_attribute[k][c].type == PLANE &&
        C->value_attribute[l][d].type == LOCATION &&
        !(C->value_attribute[i][a].type == LOCATION &&
          C->value_attribute[j][b].type == PLANE) ) {
	return( 0 );
    }
    else {
	return( 1 );
    }
}

/*
 *
 */
int
check_domain_constraint( a, time, state, type, city, number )
    int		a;			/* package number    */
    int		time;			/* max steps	     */
    int		state;			/* state of variable */
    int		type, city, number;     /* domain attributes */
{
    VALA	a_init, a_t, a_goal;

    a_init.type = LOCATION;
    a_init.city = get_orig_city( a );
    a_init.number = get_orig_loca( a );

    a_t.type = type;
    a_t.city = city;
    a_t.number = number;

    a_goal.type = LOCATION;
    a_goal.city = get_dest_city( a );
    a_goal.number = get_dest_loca( a );

    if(         state  < distance( a_init, a_t ) ||
        (time - state) < distance( a_t, a_goal ) ) {
	return( 0 );
    }

    /*
     *  Local package.
     */
    if( get_orig_city( a ) == get_dest_city( a ) ) {

	switch( type ) {

	case LOCATION:

	    /*
	     *  A local package will be at the origin and
	     *  destination locations within the city. It
	     *  may be at any other local location (for transfer
	     *  to another truck) if the number of trucks in the
	     *  city is smaller than the number of local packages
	     *  to be moved.
	     */
	    if( get_orig_city( a ) == city &&
		(get_orig_loca( a ) == number ||
		 get_dest_loca( a ) == number ||
		 possible_local_transfer( city )) ) {
		return( 1 );
	    }
	    else {
		return( 0 );
	    }

	case PLANE:
	    return( 0 );

	case TRUCK:
	    /*
	     *  A local package will only load onto a local
	     *  truck and only if the package needs to move.
	     */
	    if( get_orig_city( a ) == city &&
		get_orig_loca( a ) != get_dest_loca( a ) ) {
		return( 1 );
	    }
	    else {
		return( 0 );
	    }

	default: return( 0 );

	}
    }
    /*
     *  Non-local package.
     */
    else { /* get_orig_city( a ) != get_dest_city( a ) */

	switch( type ) {

	case LOCATION:

	    /*
	     *  A non-local package may be at an airport in
	     *  a city that is not the origination or destination
	     *  city if there is more than one plane.
	     */
	    if( get_orig_city( a ) != city &&
	        get_dest_city( a ) != city &&
	        number == get_airport_loca( city ) && 
		possible_transfer() &&
		get_num_planes() > 1 ) {
		return( 1 );
	    }
	    /*
	     *  A non-local package will be at the airport in the
	     *  origination city and at the original location. It
	     *  will only be at other locations if there is more than
	     *  one truck in the city and the package needs to move
	     *  (it didn't start at the airport).
	     */
	    else
	    if(  get_orig_city( a ) == city &&
		(number == get_airport_loca( city ) || 
		 number == get_orig_loca( a ) ||
		 (possible_local_transfer( city ) &&
		  get_orig_loca( a ) != get_airport_loca( city ))) ) {
		return( 1 );
	    }
	    /*
	     *  A non-local package will be at the airport in the
	     *  destination city and at the destination location. It
	     *  will only be at other locations if there is more than
	     *  one truck in the city and the package needs to move
	     *  (it doesn't stay at the airport).
	     */
	    else
	    if( get_dest_city( a ) == city &&
		(number == get_airport_loca( city ) ||
		 number == get_dest_loca( a ) ||
		 (possible_local_transfer( city ) &&
		  get_dest_loca( a ) != get_airport_loca( city ))) ) {
		return( 1 );
	    }
	    else {
		return( 0 );
	    }

	case PLANE:
	    return( 1 );

	case TRUCK:
	    /*
	     *  A non-local package will only load onto a truck
	     *  in the origination and destination cities and
	     *  only if the package needs to move in that city.
	     */
	    if( (get_orig_city( a ) == city &&
		 get_orig_loca( a ) != get_airport_loca( city )) ||
		(get_dest_city( a ) == city &&
		 get_dest_loca( a ) != get_airport_loca( city )) ) {
		return( 1 );
	    }
	    else {
		return( 0 );
	    }

	default: return( 0 );

	}
    }
}

