
/*
 *  CPlan solves planning problems formulated as constraint satisfaction
 *  problems.
 *
 *  Copyright (C) 1999  Peter van Beek and Xinguang Chen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.  See the GNU General Public License
 *  for more details (see the file called Copying or contact the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA).
 */

#include "logistics.h"
#include "../Share/global.h"
#include "../Share/csp.h"


int	check_domain_constraint();
/*
 *  Private routines.
 */
static void	set_domain();
static void	set_value_attributes();


/*
 *  Generate a CSP corresponding to an
 *  instance of the logistics planning problem.
 */
CSP_type *
generate( time )
    int         time;
{
    CSP_type    *C;
    int         i, j, k, a, b, c, t, t1, t2, city, v;
    int		next_con;

    C = (CSP_type *)new_calloc( 1, sizeof(CSP_type) );

    C->time = time;
    C->n = (time+1) * (get_num_packages() +
		       get_num_planes() +
		       get_num_trucks());

    /*
     *  Allocate space for n variables.
     */
    C->domain_size = new_calloc( C->n+1, sizeof(int) );
    C->variable_attribute = (VARA *)new_calloc( C->n+1, sizeof(VARA) );
    C->value_attribute = (VALA **)new_calloc( C->n+1, sizeof(VALA *) );

    /*
     *  Record the attributes of each variable.
     *  Initialize the domain of each variable.
     */
    for( t = 0; t <= time; t++ ) {
	/*
	 * Package variables
	 */
	for( i = 1; i <= get_num_packages(); i++ ) {
	    v = get_var( C, t, PACKAGE, i );
	    set_variable_attributes( C, v, PACKAGE, t, i );
	    set_domain( C, v );
	}
	/*
	 * Plane variables
	 */
	for( i = 1; i <= get_num_planes(); i++ ) {
	    v = get_var( C, t, PLANE, i );
	    set_variable_attributes( C, v, PLANE, t, i );
	    set_domain( C, v );
	}
	/*
	 * Truck variables
	 */
	for( i = 1; i <= get_num_trucks(); i++ ) {
	    v = get_var( C, t, TRUCK, i );
	    set_variable_attributes( C, v, TRUCK, t, i );
	    set_domain( C, v );
	}
    }

    /*
     *  Is the variable type visible? (1) or hidden? (0)
     */
    C->visible[PACKAGE] = 1;
    C->visible[PLANE]   = 0;
    C->visible[TRUCK]   = 0;
    C->n_visible = (time+1) * get_num_packages();


    next_con = 0;

    /*
     *  Distance constraint on the same package in different states.
     *  Enforce a lower bound on the number of time steps that would
     *  be needed to change the value of a package variable.
     */
    for( t1 = 0;    t1 <  time; t1++ )
    for( t2 = t1+1; t2 <= time; t2++ ) {
	for( a = 1; a <= get_num_packages(); a++ ) {
	    set_constraint( C, next_con, 2, DISTANCE_CONSTRAINT );
	    C->scheme[next_con][0] = get_var( C, t1, PACKAGE, a );
	    C->scheme[next_con][1] = get_var( C, t2, PACKAGE, a );
	    next_con++;
	}
    }

    /*
     *  Group outgoing packages.
     *  Two packages that both need a plane and have the same
     *  starting location should travel together from the starting
     *  location to the starting airport.
     */
    for( t = 1; t < time; t++ ) {
	for( a = 1; a <= get_num_packages(); a++ )
	for( b = a+1; b <= get_num_packages(); b++ ) {
	    if( same_origination( a, b ) &&
		needs_plane( a ) &&
		needs_plane( b ) ) {
		set_constraint( C, next_con, 2, OUTGOING_CONSTRAINT );
		C->scheme[next_con][0] = get_var( C, t, PACKAGE, a );
		C->scheme[next_con][1] = get_var( C, t, PACKAGE, b );
		next_con++;
	    }
	}
    }

    /*
     *  Group enroute packages.
     *  Two packages that both need a plane and have either
     *     a. same origination city and same destination location, or
     *     b. same origination location and same destination city
     *  should fly together in the same plane.
     */
    for( t = 1; t < time; t++ ) {
	for( a = 1; a <= get_num_packages(); a++ )
	for( b = a+1; b <= get_num_packages(); b++ ) {
	    if( needs_plane( a ) &&
	        needs_plane( b ) &&
		((same_destination( a, b ) &&
		  same_origination_city( a, b )) ||
		 (same_origination( a, b ) &&
		  same_destination_city( a, b ))) ) {
		set_constraint( C, next_con, 2, ENROUTE_CONSTRAINT );
		C->scheme[next_con][0] = get_var( C, t, PACKAGE, a );
		C->scheme[next_con][1] = get_var( C, t, PACKAGE, b );
		next_con++;
	    }
	}
    }

    /*
     *  Group incoming packages.
     *  Two packages that both need a plane and have the same
     *  destination location should travel together from the destination
     *  airport to the destination location.
     */
    for( t = 1; t < time; t++ ) {
	for( a = 1; a <= get_num_packages(); a++ )
	for( b = a+1; b <= get_num_packages(); b++ ) {
	    if( same_destination( a, b ) &&
		needs_plane( a ) &&
		needs_plane( b ) ) {
		set_constraint( C, next_con, 2, INCOMING_CONSTRAINT );
		C->scheme[next_con][0] = get_var( C, t, PACKAGE, a );
		C->scheme[next_con][1] = get_var( C, t, PACKAGE, b );
		next_con++;
	    }
	}
    }

    /*
     *  Upper bound constraint.
     *  For each outgoing package from a city, enforce an upper bound
     *  on getting the package to the airport, if it is safe to do so.
     *  For each incoming package to a city, enforce an upper bound
     *  on getting the package from the airport to its final destination,
     *  if it is safe to do so.
     */
    for( a = 1; a <= get_num_packages(); a++ ) if( needs_plane( a ) ) {
        city = get_orig_city( a );
        if( (upper_bound_outgoing( city ) +
             upper_bound_local(    city ) +
             upper_bound_incoming( city )) <= time ) {
            for( t = 1; t <= time-1; t++ ) {
                if( t >= upper_bound_outgoing( city ) ) {
                    set_constraint( C, next_con, 2, UPPER_BOUND_CONSTRAINT );
                    C->scheme[next_con][0] = get_var( C, 0, PACKAGE, a );
                    C->scheme[next_con][1] = get_var( C, t, PACKAGE, a );
                    next_con++;
                }
            }
        }
        city = get_dest_city( a );
        if( (upper_bound_outgoing( city ) +
             upper_bound_local(    city ) +
             upper_bound_incoming( city )) <= time ) {
            for( t = 1; t <= time-1; t++ ) {
                if( (time-t) >= upper_bound_incoming( city ) ) {
                    set_constraint( C, next_con, 2, UPPER_BOUND_CONSTRAINT );
                    C->scheme[next_con][0] = get_var( C, time, PACKAGE, a );
                    C->scheme[next_con][1] = get_var( C, t,    PACKAGE, a );
                    next_con++;
                }
            }
        }
    }

    /*
     *  Symmetric planes constraint (value symmetry).
     *  If the planes are all indistinguishable, then pick a package
     *  which needs a plane and ensure that the first plane the package
     *  gets loaded onto is plane 1 (arbitrary).
     */
    if( all_planes_indistinguishable() ) {
	c = 0;
	for( a = 1; a <= get_num_packages() && !c; a++ ) {
	    if( needs_plane( a ) ) {
		c = a;
	    }
	}
	if( c ) {
	    for( t = 0; t <= time-1; t++ ) {
	        set_constraint( C, next_con, 4, SYMMETRY_CONSTRAINT );
	        C->scheme[next_con][0] = get_var( C, t,   PACKAGE, c );
	        C->scheme[next_con][1] = get_var( C, t+1, PACKAGE, c );
	        C->scheme[next_con][2] = get_var( C, 0,   PACKAGE, c );
	        C->scheme[next_con][3] = get_var( C, 0,   PLANE,   1 );
	        next_con++;
	    }
	}
    }
    /*
     *  Record the arity, scheme, and relation of
     *  binary constraints on packages.
     */
    else
    if( time > 5 ) {
        for( a = 1;   a <= get_num_packages(); a++ ) if( needs_plane( a ) )
        for( b = a+1; b <= get_num_packages(); b++ ) if( needs_plane( b ) ) {
	    if( !same_origination_city( a, b ) &&
		get_orig_loca( a ) !=
		    get_airport_loca( get_orig_city( a ) ) &&
		get_orig_loca( b ) !=
		    get_airport_loca( get_orig_city( b ) ) ) {
                set_constraint( C, next_con, 2, SYMMETRY3_CONSTRAINT );
                C->scheme[next_con][0] = get_var( C, 5, PACKAGE, a );
                C->scheme[next_con][1] = get_var( C, 5, PACKAGE, b );
                next_con++;
            }
	}
    }

    /*
     *  Record the arity, scheme, and relation of
     *  binary constraints on packages.
     */
    else
    if( time > 2 ) {
        for( a = 1;   a <= get_num_packages(); a++ ) if( needs_plane( a ) )
        for( b = a+1; b <= get_num_packages(); b++ ) if( needs_plane( b ) ) {
	    if( same_origination_city( a, b ) &&
		(get_trucks_in_city( get_orig_city( a ) ) > 1) &&
	        !same_origination( a, b ) &&
		get_orig_loca( a ) !=
		    get_airport_loca( get_orig_city( a ) ) &&
		get_orig_loca( b ) !=
		    get_airport_loca( get_orig_city( b ) ) ) {
                set_constraint( C, next_con, 2, SYMMETRY4_CONSTRAINT );
                C->scheme[next_con][0] = get_var( C, 2, PACKAGE, a );
                C->scheme[next_con][1] = get_var( C, 2, PACKAGE, b );
                next_con++;
            }
	}
    }

    /*
     *  Symmetry constraint, earliest possible pickup.
     *  If there are two packages, a and b, at a location, and
     *  package a gets loaded onto a plane and package b does
     *  not get loaded onto that plane, then package b should
     *  not get loaded onto that plane in some future state.
     */
    for( t1 = 0;    t1 <= time-2; t1++ )
    for( t2 = t1+2; t2 <= time;   t2++ ) {
	for( a = 1; a <= get_num_packages(); a++ )
	for( b = 1; b <= get_num_packages(); b++ ) if( a != b ) {
		j = get_var( C, 0, PACKAGE, b );
		k = get_var( C, time, PACKAGE, a );
		if( same_origination_city( a, b ) &&
		    needs_plane( a ) &&
		    needs_plane( b ) ) {
		    set_constraint( C, next_con, 5, SYMMETRY2_CONSTRAINT );
		    C->scheme[next_con][0] = get_var( C, t1,   PACKAGE, a );
		    C->scheme[next_con][1] = get_var( C, t1+1, PACKAGE, a );
		    C->scheme[next_con][2] = get_var( C, t1,   PACKAGE, b );
		    C->scheme[next_con][3] = get_var( C, t1+1, PACKAGE, b );
		    C->scheme[next_con][4] = get_var( C, t2,   PACKAGE, b );
		    next_con++;
		}
		if( (get_dest_city( a ) == get_orig_city( b )) &&
		    needs_plane( a ) &&
		    needs_plane( b ) ) {
		    set_constraint( C, next_con, 5, SYMMETRY2_CONSTRAINT );
		    C->scheme[next_con][0] = get_var( C, t1,   PACKAGE, a );
		    C->scheme[next_con][1] = get_var( C, t1+1, PACKAGE, a );
		    C->scheme[next_con][2] = get_var( C, t1,   PACKAGE, b );
		    C->scheme[next_con][3] = get_var( C, t1+1, PACKAGE, b );
		    C->scheme[next_con][4] = get_var( C, t2,   PACKAGE, b );
		    next_con++;
		}
	}
    }

    /*
     *  Record the arity, scheme, and relation of a
     *  4-ary constraints on packages and planes.
     */
    for( t = 0; t < time; t++ ) {
	for( a = 1; a <= get_num_packages(); a++ )
	for( b = 1; b <= get_num_planes();   b++ ) {
	    set_constraint( C, next_con, 4, SUCCESSFUL_LOAD_UNLOAD_CONSTRAINT );
	    C->scheme[next_con][0] = get_var( C, t,   PACKAGE, a );
	    C->scheme[next_con][1] = get_var( C, t,   PLANE,   b );
	    C->scheme[next_con][2] = get_var( C, t+1, PACKAGE, a );
	    C->scheme[next_con][3] = get_var( C, t+1, PLANE,   b );
	    next_con++;
	}
    }

    /*
     *  Record the arity, scheme, and relation of a
     *  4-ary constraints on packages and trucks.
     */
    for( t = 0; t < time; t++ ) {
	for( a = 1; a <= get_num_packages(); a++ )
	for( b = 1; b <= get_num_trucks();   b++ ) {
	    set_constraint( C, next_con, 4, SUCCESSFUL_LOAD_UNLOAD_CONSTRAINT );
	    C->scheme[next_con][0] = get_var( C, t,   PACKAGE, a );
	    C->scheme[next_con][1] = get_var( C, t,   TRUCK,   b );
	    C->scheme[next_con][2] = get_var( C, t+1, PACKAGE, a );
	    C->scheme[next_con][3] = get_var( C, t+1, TRUCK,   b );
	    next_con++;
	}
    }

    /*
     *  Record the arity, scheme, and relation of a
     *  3-ary constraint on packages.
     */
    for( t  =    0; t  <= time; t ++ )
    for( t1 = t +1; t1 <= time; t1++ )
    for( t2 = t1+1; t2 <= time; t2++ ) {
	for( a = 1; a <= get_num_packages(); a++ ) {
	    set_constraint( C, next_con, 3, USELESS_ACTIONS_CONSTRAINT );
	    C->scheme[next_con][0] = get_var( C, t,  PACKAGE, a );
	    C->scheme[next_con][1] = get_var( C, t1, PACKAGE, a );
	    C->scheme[next_con][2] = get_var( C, t2, PACKAGE, a );
	    next_con++;
	}
    }

    /*
     *  Force parallelism constraint.
     *  Record the arity, scheme, and relation of a
     *  4-ary constraints on packages and packages.
     */
    for( t = 0; t <= time-2; t++ ) {
	for( a = 1; a <= get_num_packages(); a++ )
	for( b = 1; b <= get_num_packages(); b++ ) {
	    set_constraint( C, next_con, 4, FORCE_PARALLELISM_CONSTRAINT );
	    C->scheme[next_con][0] = get_var( C, t,   PACKAGE, a );
	    C->scheme[next_con][1] = get_var( C, t+1, PACKAGE, a );
	    C->scheme[next_con][2] = get_var( C, t+1, PACKAGE, b );
	    C->scheme[next_con][3] = get_var( C, t+2, PACKAGE, b );
	    next_con++;
	}
    }

    /*
     *  Record the arity, scheme, and relation of a
     *  4-ary constraint on packages.
     */
    for( t1 = 0;    t1 <  time; t1++ )
    for( t2 = t1+1; t2 <= time; t2++ ) {
	for( a = 1;   a <= get_num_packages(); a++ )
	for( b = a+1; b <= get_num_packages(); b++ ) {
	    set_constraint( C, next_con, 4, DISTANCE2_CONSTRAINT );
	    C->scheme[next_con][0] = get_var( C, t1, PACKAGE, a );
	    C->scheme[next_con][1] = get_var( C, t2, PACKAGE, a );
	    C->scheme[next_con][2] = get_var( C, t1, PACKAGE, b );
	    C->scheme[next_con][3] = get_var( C, t2, PACKAGE, b );
	    next_con++;
	}
    }

    for( a = 1; a <= get_num_packages(); a++ ) {
	if( force_unload( a ) ) {
	    for( t = 0; t <= time-3; t++ ) {
		set_constraint( C, next_con, 4, FORCE_MOVE_CONSTRAINT );
	        C->scheme[next_con][0] = get_var( C, t,   PACKAGE, a );
	        C->scheme[next_con][1] = get_var( C, t+1, PACKAGE, a );
	        C->scheme[next_con][2] = get_var( C, t+2, PACKAGE, a );
	        C->scheme[next_con][3] = get_var( C, t+3, PACKAGE, a );
		next_con++;
	    }
        }
    }

    C->m = next_con;

    /*
     *  When should the constraint be propagated?
     *
     *  C->checkable[r] = u means that constraint type r is propagated
     *          only when the number of uninstantiated variables in the
     *          scheme of the constraint is less than or equal to u.
     *          Constraint propagation then potentially reduces the
     *          domains of each of the uninstantiated variables.
     *  
     *  In particular,
     *    == 1    means only forward checking done on this constraint
     *    >= 2    means arc consistency checking done on this constraint
     *    == HUGE means constraint is always propagated
     */
    C->checkable[DISTANCE_CONSTRAINT] = HUGE;
    C->checkable[DISTANCE2_CONSTRAINT] = 1;
    C->checkable[SUCCESSFUL_LOAD_UNLOAD_CONSTRAINT] = HUGE;
    C->checkable[USELESS_ACTIONS_CONSTRAINT] = 1;
    C->checkable[FORCE_PARALLELISM_CONSTRAINT] = 1;
    C->checkable[OUTGOING_CONSTRAINT] = HUGE;
    C->checkable[INCOMING_CONSTRAINT] = HUGE;
    C->checkable[ENROUTE_CONSTRAINT] = HUGE;
    C->checkable[UPPER_BOUND_CONSTRAINT] = 1;
    C->checkable[SYMMETRY_CONSTRAINT] = HUGE;
    C->checkable[SYMMETRY2_CONSTRAINT] = 1;
    C->checkable[SYMMETRY3_CONSTRAINT] = HUGE;
    C->checkable[SYMMETRY4_CONSTRAINT] = HUGE;
    C->checkable[FORCE_MOVE_CONSTRAINT] = HUGE;

    construct_index( C );

    return( C );
}

int
get_var( C, state, type, number )
    CSP_type    *C;
    int		state, type, number;
{
    int		i;

    i = state * (get_num_packages() +
		 get_num_planes() +
		 get_num_trucks());

    if( type == PACKAGE ) {
	i += number;
    }
    else
    if( type == PLANE ) {
	i += get_num_packages() + number;
    }
    else
    if( type == TRUCK ) {
	i += get_num_packages() + get_num_planes() + number;
    }
    return( i );
}

static void
set_value_attributes( C, v, a, type, city, number )
    CSP_type    *C;
    int		v, a;
    int		type, city, number;
{
    C->value_attribute[v][a].type = type;
    C->value_attribute[v][a].city = city;
    C->value_attribute[v][a].number = number;
}

/*
 *  Initialize the domain of a variable.
 *  Record the attributes of each value.
 */
static void
set_domain( C, v )
    CSP_type    *C;
    int		v;
{
    int		state, type, a, j, k;
    int		next_val;

    get_variable_attributes( C, v, &type, &state, &a );

    /*
     * Package variables.
     * Domains are all locations, all planes, all trucks.
     */
    if( type == PACKAGE ) {
	/*
	 *  Set initial state.
	 */
	if( (state == 0) || (get_orig_city( a ) == get_dest_city( a ) &&
			     get_orig_loca( a ) == get_dest_loca( a )) ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
	    set_value_attributes( C, v, 0, LOCATION,
		get_orig_city( a ), get_orig_loca( a ) );
	    C->domain_size[v] = 1;
	}
	/*
	 *  Set goal state.
	 */
	else
	if( state == C->time ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
	    set_value_attributes( C, v, 0, LOCATION,
		get_dest_city( a ), get_dest_loca( a ) );
	    C->domain_size[v] = 1;
	}
	/*
	 *  Set general state.
	 */
	else {
            C->value_attribute[v] = (VALA *)
		new_calloc( get_loc_per_city()*get_num_cities() +
			    get_num_planes() +
			    get_num_trucks(), sizeof(VALA) );
	    next_val = 0;

	    for( j = 1; j <= get_num_cities(); j++ ) {
		for( k = 1; k <= get_loc_per_city(); k++ ) {
		  if( check_domain_constraint( a, C->time, state, LOCATION, j, k ) ) {
		    set_value_attributes( C, v, next_val, LOCATION, j, k );
		    next_val++;
		  }
		}
	    }

	    for( j = 1; j <= get_num_planes(); j++ ) {
	      if( check_domain_constraint( a, C->time, state, PLANE, 0, j ) ) {
		set_value_attributes( C, v, next_val, PLANE, 0/*unused*/, j );
		next_val++;
	      }
	    }

	    for( j = 1; j <= get_num_trucks(); j++ ) {
		if( check_domain_constraint( a, C->time, state, TRUCK,
			get_truck_orig_city( j ), j ) ) {
		    set_value_attributes( C, v, next_val, TRUCK,
			get_truck_orig_city( j ), j );
		    next_val++;
		}
	    }
	    C->domain_size[v] = next_val;
	}
    }
    else
    /*
     * Plane variables.
     * Domains are locations which are airports (one per city).
     */
    if( type == PLANE ) {
	/*
	 *  Set initial state.
	 */
	if( state == 0 ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
	    set_value_attributes( C, v, 0, LOCATION,
		get_plane_orig_city( a ), get_plane_orig_loca( a ) );
	    C->domain_size[v] = 1;
	}
	/*
	 *  Set general state.
	 */
	else {
            C->value_attribute[v] = (VALA *)
		new_calloc( get_num_cities(), sizeof(VALA) );
	    next_val = 0;
	    for( j = 1; j <= get_num_cities(); j++ ) {
		set_value_attributes( C, v, next_val,
			LOCATION, j, get_airport_loca( j ) );
		next_val++;
	    }
	    C->domain_size[v] = next_val;
	}
    }
    else
    /*
     * Truck variables.
     * Domains are locations within a city.
     */
    if( type == TRUCK ) {
	/*
	 *  Set initial state.
	 */
	if( state == 0 ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
	    set_value_attributes( C, v, 0, LOCATION,
		get_truck_orig_city( a ), get_truck_orig_loca( a ) );
	    C->domain_size[v] = 1;
	}
	/*
	 *  Set general state.
	 */
	else {
            C->value_attribute[v] = (VALA *)
		new_calloc( get_loc_per_city(), sizeof(VALA) );
	    next_val = 0;
	    for( j = 1; j <= get_loc_per_city(); j++ ) {
		set_value_attributes( C, v, next_val, LOCATION,
			get_truck_orig_city( a ), j );
		next_val++;
	    }
	    C->domain_size[v] = next_val;
	}
    }
}

