
/*
 *  CPlan solves planning problems formulated as constraint satisfaction
 *  problems.
 *
 *  Copyright (C) 1999  Peter van Beek and Xinguang Chen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.  See the GNU General Public License
 *  for more details (see the file called Copying or contact the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA).
 */

#include <math.h>
#include "logistics.h"
#include "../Share/global.h"
#include "../Share/csp.h"


static I_type *L;

static void   city_info();
static void   reduce_packages();
static void   reduce_planes();
static void   reduce_trucks();
static int    plane_must_move_or_wait();	/* one plane  */
static int    truck_must_move_or_wait();	/* one truck  */


void
get_instance( problem )
    int		problem;
{
    int		i;
    int		package, plane, truck, city, location;
    char	name[30];
    FILE	*fp;

    if( problem < 10 ) sprintf( name, "./Problems/p0%1d", problem );
    else	       sprintf( name, "./Problems/p%2d",  problem );

    if( (fp = fopen( name, "r" )) == NULL ) {
	printf( "Could not open file %s\n", name );
	exit( 0 );
    }

    L = (I_type *)new_calloc( 1, sizeof(I_type) );

    /*
     *  Get and set parameters.
     */
    L->problem = problem;
    fscanf( fp, "Parameters:\n" );
    fscanf( fp, "\tname: %s\n", &name[0] );
    fscanf( fp, "\tpackages: %d\n", &(L->packages) );
    fscanf( fp, "\tplanes: %d\n", &(L->planes) );
    fscanf( fp, "\tcities: %d\n", &(L->cities) );
    fscanf( fp, "\tlocations_per_city: %d\n", &(L->locations_per_city) );
    fscanf( fp, "\ttrucks: %d\n", &(L->trucks) );
    fscanf( fp, "\n" );

    /*
     *  Allocate space to store the initial and goal states.
     */
    L->orig_city = new_calloc( L->packages+1, sizeof(int) );
    L->orig_loca = new_calloc( L->packages+1, sizeof(int) );
    L->dest_city = new_calloc( L->packages+1, sizeof(int) );
    L->dest_loca = new_calloc( L->packages+1, sizeof(int) );

    L->plane_orig_city = new_calloc( L->planes+1, sizeof(int) );
    L->plane_orig_loca = new_calloc( L->planes+1, sizeof(int) );
    L->truck_orig_city = new_calloc( L->trucks+1, sizeof(int) );
    L->truck_orig_loca = new_calloc( L->trucks+1, sizeof(int) );

    L->airport_loca    = new_calloc( L->cities+1, sizeof(int) );
    L->outgoing        = new_calloc( L->cities+1, sizeof(int) );
    L->local           = new_calloc( L->cities+1, sizeof(int) );
    L->incoming        = new_calloc( L->cities+1, sizeof(int) );
    L->trucks_in_city  = new_calloc( L->cities+1, sizeof(int) );

    L->package_name = new_calloc( L->packages+1, sizeof(int) );
    L->plane_name   = new_calloc( L->planes+1,   sizeof(int) );
    L->truck_name   = new_calloc( L->trucks+1,   sizeof(int) );

    fscanf( fp, "Initial:\n" );
    for( i = 1; i <= L->cities; i++ ) {
	fscanf( fp, "\t(airport city%d-%d)\n", &city, &location );
	L->airport_loca[city] = location;
    }
    for( i = 1; i <= L->packages; i++ ) {
	fscanf( fp, "\t(at package%d city%d-%d)\n",
		&package, &city, &location );
	L->orig_city[package] = city;
	L->orig_loca[package] = location;
	L->package_name[package] = package;
    }
    for( i = 1; i <= L->planes; i++ ) {
	fscanf( fp, "\t(at plane%d city%d-%d)\n",
		&plane, &city, &location );
	L->plane_orig_city[plane] = city;
	L->plane_orig_loca[plane] = location;
	L->plane_name[plane] = plane;
    }
    for( i = 1; i <= L->trucks; i++ ) {
	fscanf( fp, "\t(at truck%d city%d-%d)\n",
		&truck, &city, &location );
	L->truck_orig_city[truck] = city;
	L->truck_orig_loca[truck] = location;
	L->trucks_in_city[city]++;
	L->truck_name[truck] = truck;
    }
    fscanf( fp, "\n" );

    for( i = 1; i <= L->packages; i++ ) {
	L->dest_city[i] = L->orig_city[i];
	L->dest_loca[i] = L->orig_loca[i];
    }

    fscanf( fp, "Goal:\n" );
    while( fscanf( fp, "\t(at package%d city%d-%d)\n",
		&package, &city, &location ) > 0 ) {
	L->dest_city[package] = city;
	L->dest_loca[package] = location;
    }

    fclose( fp );

    city_info();
    reduce_packages();
    reduce_planes();
    reduce_trucks();
}

static void
reduce_packages()
{
    int i, j;

    i = 0;
    for( j = 1; j <= L->packages; j++ ) {
	if( (L->dest_city[j] != L->orig_city[j]) ||
	    (L->dest_loca[j] != L->orig_loca[j]) ) {
	    i++;
	    L->orig_city[i] = L->orig_city[j];
	    L->orig_loca[i] = L->orig_loca[j];
	    L->dest_city[i] = L->dest_city[j];
	    L->dest_loca[i] = L->dest_loca[j];
	    L->package_name[i] = L->package_name[j];
        }
    }
    L->packages = i;
}

static void
reduce_planes()
{
    int i, j, max_planes_needed, excess;

    max_planes_needed = 0;
    for( i = 1; i <= L->packages; i++ )
	if( needs_plane( i ) ) {
	    max_planes_needed++;
	}

    if( L->planes > max_planes_needed ) {
	excess = L->planes - max_planes_needed;
	i = 0;
	for( j = 1; j <= L->planes; j++ ) {
	    if( excess && plane_must_move_or_wait( j ) ) {
		/* skip this plane */
		excess--;
	    }
	    else {
		i++;
		L->plane_orig_city[i] = L->plane_orig_city[j];
		L->plane_orig_loca[i] = L->plane_orig_loca[j];
		L->plane_name[i] = L->plane_name[j];
	    }
	}
	L->planes = i;
    }
}

static void
reduce_trucks()
{
    int i, j, city, max_trucks_needed, excess;

    for( city = 1; city <= L->cities; city++ ) {
	max_trucks_needed = L->outgoing[city] +
			    L->local[city] +
			    L->incoming[city];

	if( L->trucks_in_city[city] > max_trucks_needed ) {
	    excess = L->trucks_in_city[city] - max_trucks_needed;
	    i = 0;
	    for( j = 1; j <= L->trucks; j++ ) {
		if( excess &&
		    L->truck_orig_city[j] == city &&
		    truck_must_move_or_wait( j ) ) {
		    /* skip this truck */
		    excess--;
		    L->trucks_in_city[city]--;
		}
		else {
		    i++;
		    L->truck_orig_city[i] = L->truck_orig_city[j];
		    L->truck_orig_loca[i] = L->truck_orig_loca[j];
		    L->truck_name[i] = L->truck_name[j];
		}
	    }
	    L->trucks = i;
	}
    }
}

void
print_instance()
{
    printf( "Parameters:\n" );
    printf("\tproblem: %d\n", L->problem );
    printf("\tpackages: %d\n", L->packages );
    printf("\tplanes: %d\n", L->planes );
    printf("\tcities: %d\n", L->cities );
    printf("\tlocations_per_city: %d\n", L->locations_per_city );
    printf("\ttrucks: %d\n", L->trucks );
    printf("\tlower bound: %d\n", get_lower_bound() );
    printf("\tupper bound: %d\n", get_upper_bound() );
}

void
dump_instance()
{
    int		i;

    printf( "\nInitial:\n" );
    for( i = 1; i <= L->cities; i++ ) {
	printf( "\t(airport city%d-%d)\n", i, L->airport_loca[i] );
    }
    for( i = 1; i <= L->packages; i++ ) {
	printf( "\t(at package%d city%d-%d)\n",
		i, L->orig_city[i], L->orig_loca[i] );
    }
    for( i = 1; i <= L->planes; i++ ) {
	printf( "\t(at plane%d city%d-%d)\n",
		i, L->plane_orig_city[i], L->plane_orig_loca[i] );
    }
    for( i = 1; i <= L->trucks; i++ ) {
	printf( "\t(at truck%d city%d-%d)\n",
		i, L->truck_orig_city[i], L->truck_orig_loca[i] );
    }

    printf( "\nGoal:\n" );
    for( i = 1; i <= L->packages; i++ ) {
	if( (L->dest_city[i] != L->orig_city[i]) ||
	    (L->dest_loca[i] != L->orig_loca[i]) )
	    printf( "\t(at package%d city%d-%d)\n",
		i, L->dest_city[i], L->dest_loca[i] );
    }
    printf( "\n" );
}

int get_num_packages()	{ return( L->packages ); }
int get_num_planes()	{ return( L->planes ); }
int get_num_cities()	{ return( L->cities ); }
int get_loc_per_city()  { return( L->locations_per_city ); }
int get_num_trucks()	{ return( L->trucks ); }

int get_package_name( int i ) { return( L->package_name[i] ); }
int get_plane_name(   int i ) { return( L->plane_name[i]   ); }
int get_truck_name(   int i ) { return( L->truck_name[i]   ); }

int get_orig_city( int i ) { return( L->orig_city[i] ); }
int get_orig_loca( int i ) { return( L->orig_loca[i] ); }
int get_dest_city( int i ) { return( L->dest_city[i] ); }
int get_dest_loca( int i ) { return( L->dest_loca[i] ); }

int get_plane_orig_city( int i ) { return( L->plane_orig_city[i] ); }
int get_plane_orig_loca( int i ) { return( L->plane_orig_loca[i] ); }
int get_truck_orig_city( int i ) { return( L->truck_orig_city[i] ); }
int get_truck_orig_loca( int i ) { return( L->truck_orig_loca[i] ); }
int get_trucks_in_city(  int i ) { return( L->trucks_in_city[i] ); }
int get_airport_loca(    int i ) { return( L->airport_loca[i]    ); }

int possible_transfer()
{
    return( L->planes < L->packages );
}

int possible_local_transfer( int i )
{
    return( L->trucks_in_city[i] < L->local[i] );
}

static int
possibly_share_a_plane( j, track )
    int         j, *track;
{
    int		i;

    for( i = 1; i < j; i++ ) if( track[i] ) {
	if( L->orig_city[i] == L->orig_city[j] &&
	    L->dest_city[i] == L->dest_city[j] ) {
	    return( 1 );
	}
    }

    return( 0 );
}

int get_lower_bound()
{
    int		j, required, track[50];

    for( j = 1; j <= L->packages; j++ ) {
	track[j] = 0;
    }

    required = 0;
    for( j = 1; j <= L->packages; j++ ) {
	if( needs_plane( j ) &&
            L->orig_loca[j] != L->airport_loca[L->orig_city[j]] &&
            L->dest_loca[j] != L->airport_loca[L->dest_city[j]] ) {

            /*
             *  Package necessarily needs a plane.
             *  Check whether it could fly with another package
             *  because the source and destination cities are the same.
             */
            if( !possibly_share_a_plane( j, track ) ) {
                required++;
                track[j] = 1;
            }
        }
    }

    if( required == 0 ) {
	return( 1 );
    }
    else
    if( required <= L->planes ) {
	return( 9 );
    }
    else {
	return( 11 );
    }

}

int
get_upper_bound()
{
    int		city, t, max_to_airport, max_from_airport;

    max_to_airport = 0;
    max_from_airport = 0;
    for( city = 1; city <= L->cities; city++ ) {
	t = upper_bound_outgoing( city );
	if( t > max_to_airport )
	    max_to_airport = t;
	t = upper_bound_incoming( city );
	if( t > max_from_airport )
	    max_from_airport = t;
    }

    return( max_to_airport +
	    (4 * (int)ceil( (double)L->packages /
			    (double)L->planes ))  +
	    max_from_airport
          );
}

int
upper_bound_outgoing( int city )
{
    int a, b, flag, exists;

    /*
     *  If all outgoing packages in city have a truck
     *  at their location, then 3 steps for some of the packages
     */
    flag = 1;
    for( a = 1; a <= L->packages && flag; a++ ) {
        if( L->orig_city[a] == city && needs_plane( a ) ) {
	    exists = 0;
	    for( b = 1; b <= L->trucks && flag; b++ ) {
		if( L->truck_orig_city[b] == city &&
		    L->truck_orig_loca[b] == L->orig_loca[a] ) {
		    exists = 1;
		}
	    }
	    if( !exists ) {
		flag = 0;
	    }
        }
    }

    if( !flag ) {
        return( 4 * (int)ceil( (double)L->outgoing[city] /
			       (double)L->trucks_in_city[city] ) );
    }
    else {
        if( L->trucks_in_city[city] >= L->outgoing[city] ) {
	    return( 3 );
        }
        else { /* L->trucks_in_city[city] < L->outgoing[city] */
	    return( 3 +
		    4 * (int)ceil( (double)(L->outgoing[city] -
					    L->trucks_in_city[city]) /
			           (double)L->trucks_in_city[city] )
	          );
        }
    }
}

int
upper_bound_local( int city )
{
    if( L->local[city] == 0 ) {
	return( 0 );
    }
    else
    if( L->incoming[city] == 0 ) {
	return( 4 * (int)ceil( (double)L->local[city] /
			       (double)L->trucks_in_city[city] ) );
    }
    else {
	return( 4 * (int)ceil( (double)L->local[city] /
			       (double)L->trucks_in_city[city] ) + 1 );
					/* L->incoming[city] != 0,   */
					/* add 1 for move to airport */
    }
}

int
upper_bound_incoming( int city )
{
    if( L->incoming[city] == 0 ) {
	return( 0 );
    }
    else
    if( L->trucks_in_city[city] >= L->incoming[city] ) {
	return( 3 );
    }
    else { /* L->trucks_in_city[city] < L->incoming[city] */
	return( 3 +
		4 * (int)ceil( (double)(L->incoming[city] -
					L->trucks_in_city[city]) /
			       (double)L->trucks_in_city[city] )
	      );
    }
}

int
needs_plane( a )
    int		a;
{
    return( L->orig_city[a] != L->dest_city[a] );
}

int
same_origination( a, b )
    int		a, b;
{
    return( L->orig_city[a] == L->orig_city[b] &&
	    L->orig_loca[a] == L->orig_loca[b] );
}

int
same_origination_city( a, b )
    int		a, b;
{
    return( L->orig_city[a] == L->orig_city[b] );
}

int
same_destination( a, b )
    int		a, b;
{
    return( L->dest_city[a] == L->dest_city[b] &&
	    L->dest_loca[a] == L->dest_loca[b] );
}

int
same_destination_city( a, b )
    int		a, b;
{
    return( L->dest_city[a] == L->dest_city[b] );
}

/*
 *  Is there a package (that needs a plane) at the
 *  same airport as the plane in the initial state?
 */
static int
plane_must_move_or_wait( a )
    int		a;
{
    int         b, city, flag;

    flag = 1;
    city = L->plane_orig_city[a];
    for( b = 1; b <= L->packages && flag; b++ ) {
        if( L->orig_city[b] == city &&
            L->orig_loca[b] == L->airport_loca[city] &&
            needs_plane( b ) ) {
            flag = 0;
        }
    }

    return( flag );
}

/*
 *  Is there a package that needs some truck to the
 *  airport at the same location as the given truck
 *  in the initial state?
 */
static int
truck_must_move_or_wait( a )
    int		a;
{
    int         b, city, loca, flag;

    flag = 1;
    city = L->truck_orig_city[a];
    loca = L->truck_orig_loca[a];
    for( b = 1; b <= L->packages && flag; b++ ) {
        if( L->orig_city[b] == city &&
            L->orig_loca[b] == loca &&
            needs_plane( b ) &&
	    L->orig_loca[b] != L->airport_loca[city] ) {
            flag = 0;
        }
    }

    return( flag );
}

/*
 *  Two trucks are indistinguishable if:
 *      both trucks are at the city and location in the initial state, or
 *      both trucks have to move or wait to pickup a first package.
 */
int
indistinguishable_trucks( a, b )
    int		a, b;
{
    return( (L->truck_orig_city[a] == L->truck_orig_city[b] &&
	     L->truck_orig_loca[a] == L->truck_orig_loca[b]) ||
	    (truck_must_move_or_wait( a ) && truck_must_move_or_wait( b ))
	  );
}

/*
 *  Two planes are indistinguishable if:
 *      both planes are at the same airport in the initial state, or
 *      both planes have to move or wait to pickup a first package.
 */
int
indistinguishable_planes( a, b )
    int		a, b;
{
    return( (L->plane_orig_city[a] == L->plane_orig_city[b]) ||
	    (plane_must_move_or_wait( a ) && plane_must_move_or_wait( b ))
	  );
}

int
all_planes_indistinguishable()
{
    int         a, b;

    for( a = 1; a <= L->planes; a++ ) {
	for( b = a+1; b <= L->planes; b++ ) {
	    if( !indistinguishable_planes( a, b ) ) {
		return( 0 );
	    }
	}
    }

    return( 1 );
}

static void
city_info()
{
    int         a, l, city;

    int locations_outgoing[20];
    int locations_incoming[20];

    for( city = 1; city <= L->cities; city++ ) {

	L->outgoing[city] = 0;
	L->local[city] = 0;
	L->incoming[city] = 0;

	for( l = 1; l <= L->locations_per_city; l++ ) {
	    /* locations within the city that have a package
		destined for the airport */
	    locations_outgoing[l] = 0;
	    /* locations within the city that have a package
		coming from the airport */
	    locations_incoming[l] = 0;
	}

	for( a = 1; a <= L->packages; a++ ) {
	    /*
	     *  Local package.
	     */
	    if( L->orig_city[a] == city &&
		L->dest_city[a] == city ) {

		if( L->orig_loca[a] != L->dest_loca[a] ) {
		    L->local[city]++;
		}
	    }
	    /*
	     *  Outgoing package.
	     */
	    else
	    if( L->orig_city[a] == city &&
		L->dest_city[a] != city ) {

		if( L->orig_loca[a] == L->airport_loca[city] ) {
		    /* already at airport, nothing more to do */
		}
		else
		if( !locations_outgoing[L->orig_loca[a]] ) {
		    /* new location within city */
		    L->outgoing[city]++;
		    locations_outgoing[L->orig_loca[a]] = 1;
		}
	    }
	    /*
	     *  Incoming package.
	     */
	    else
	    if( L->orig_city[a] != city &&
		L->dest_city[a] == city ) {

		if( L->dest_loca[a] == L->airport_loca[city] ) {
		    /* destination is airport, nothing more to do */
		}
		else
		if( !locations_incoming[L->dest_loca[a]] ) {
		    /* new location within city */
		    L->incoming[city]++;
		    locations_incoming[L->dest_loca[a]] = 1;
		}
	    }

	}
    }
}

int
force_unload( int p )
{
    int		j, in[50], out[50];

    for( j = 1; j <= L->cities; j++ ) {
	in[j] = 0;
	out[j] = 0;
    }

    for( j = 1; j <= L->packages; j++ ) if( needs_plane( j ) ) {
	out[L->orig_city[j]]++;
	in[L->dest_city[j]]++;
    }

    if( needs_plane( p ) &&
	(L->orig_loca[p] != L->airport_loca[L->orig_city[p]]) &&
		(L->trucks_in_city[L->orig_city[p]] >=
	L->incoming[L->orig_city[p]]) &&
	out[L->orig_city[p]] == 1 &&
	in[L->dest_city[p]] == 1 ) {

	return( 1 );
    }
    else {
	return( 0 );
    }
}

