
/*
 *  CPlan solves planning problems formulated as constraint satisfaction
 *  problems.
 *
 *  Copyright (C) 1999  Peter van Beek and Xinguang Chen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.  See the GNU General Public License
 *  for more details (see the file called Copying or contact the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA).
 */

#include "mystery.h"
#include "../Share/csp.h"
#include "../Share/global.h"

static int      check_cargo_distance_constraint();
static int      check_vehicle_distance_constraint();
static int      check_fuel_distance_constraint();
static int      check_action_load_unload_constraint();
static int      check_action_move_constraint();
static int      check_vehicle_capacity_constraint();
static int      check_single_move_constraint();
static int      check_action_squirt_constraint();
static int      check_single_squirt_move_constraint();

/*
 *  Check the given values against the constraint.
 */
int
check_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    checks += C->arity[r];

    switch( C->relation[r] ) {

    case DISTANCE_C_LOC_CONSTRAINT:
        return( check_cargo_distance_constraint( C, r, solution ) );

    case DISTANCE_V_LOC_CONSTRAINT:
        return( check_vehicle_distance_constraint( C, r, solution ) );

    case DISTANCE_N_FUEL_CONSTRAINT:
        return( check_fuel_distance_constraint( C, r, solution ) );

    case ACTION_LOAD_UNLOAD_CONSTRAINT:
        return( check_action_load_unload_constraint( C, r, solution ) );

    case ACTION_MOVE_CONSTRAINT:
        return( check_action_move_constraint( C, r, solution ) );

    case VEHICLE_CAPACITY_CONSTRAINT:
        return( check_vehicle_capacity_constraint( C, r, solution ) );

    case SINGLE_MOVE_CONSTRAINT:
        return( check_single_move_constraint( C, r, solution ) );

    case ACTION_SQUIRT_CONSTRAINT:
        return( check_action_squirt_constraint( C, r, solution ) );

    case SINGLE_SQUIRT_MOVE_CONSTRAINT:
        return( check_single_squirt_move_constraint( C, r, solution ) );

    default:
        printf("Error in check_constraint.\n");
        exit(6);
    }
}

/*
 *  Determine a lower bound on the number of time steps that would
 *  be needed to change the value of a cargo variable.
 */
static int
cargo_distance( a, a_new )
    VALA         a, a_new;
{
    if( a.type == VEHICLE && a_new.type == VEHICLE ) {
        if( a.number == a_new.number ) return( 0 );
        else                           return( 2 );
    }
    else
    if( a.type == VEHICLE && a_new.type == NODE ) {
        return( 1 );
    }
    else
    if( a.type == NODE && a_new.type == VEHICLE ) {
        return( 1 );
    }
    else
    if( a.type == NODE && a_new.type == NODE ) {
        if( a.number == a_new.number )
            return( 0 );
        else                    
            return( 2 + sp_node_distance( a.number, a_new.number ) );
    }
    else {
        printf("Error in cargo_distance.\n");
        exit(6);
    }
}

static int
check_cargo_distance_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;

    i = C->scheme[r][0];  /* cargo(t1) */
    j = C->scheme[r][1];  /* cargo(t2), 0 <= t1 < t2 <= time */

    a = solution[i];
    b = solution[j];

    /*
     *  actual distance >= lower bound on distance
     */
    return(
        (C->variable_attribute[j].state - C->variable_attribute[i].state) >=
        cargo_distance( C->value_attribute[i][a], C->value_attribute[j][b] )
    );
}

/*
 *  Determine a lower bound on the number of time steps that would
 *  be needed to change the value of a vehicle variable.
 */
static int
vehicle_distance( a, a_new )
    VALA         a, a_new;
{
    if( a.number == a_new.number ) {
        return( 0 );
    }
    else {
        return( sp_node_distance( a.number, a_new.number ) );
    }
}

static int
check_vehicle_distance_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;

    i = C->scheme[r][0];  /* vehicle(t1) */
    j = C->scheme[r][1];  /* vehicle(t2), 0 <= t1 < t2 <= time */

    a = solution[i];
    b = solution[j];

    /*
     *  actual distance >= lower bound on distance
     */
    return(
        (C->variable_attribute[j].state - C->variable_attribute[i].state) >=
        vehicle_distance( C->value_attribute[i][a], C->value_attribute[j][b] )
    );
}

/*
 *  Determine a lower bound on the number of time steps that would
 *  be needed to change the value of a fuel variable.
 */
static int
fuel_distance( a, a_new )
    VALA         a, a_new;
{
    if( a.number < a_new.number ) {
        if( squirt_action() && a.number == 0 && a_new.number == 1 )
            return( 1 );
        else
            return( HUGE );
    }
    else
    if( a.number == a_new.number ) {
        return( 0 );
    }
    else /* a.number > a_new.number */ {
        return( a.number - a_new.number );
    }
}

static int
check_fuel_distance_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;

    i = C->scheme[r][0];  /* fuel(t1) */
    j = C->scheme[r][1];  /* fuel(t2), 0 <= t1 < t2 <= time */

    a = solution[i];
    b = solution[j];

    /*
     *  actual distance >= lower bound on distance
     */
    return(
        (C->variable_attribute[j].state - C->variable_attribute[i].state) >=
        fuel_distance( C->value_attribute[i][a], C->value_attribute[j][b] )
    );
}

static int
check_action_load_unload_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j, k, l;
    int         a, b, c, d;

    i = C->scheme[r][0]; /* cargo(t)   */
    j = C->scheme[r][1]; /* vehicle(t)   */
    k = C->scheme[r][2]; /* cargo(t+1) */
    l = C->scheme[r][3]; /* vehicle(t+1) */

    a = solution[i];
    b = solution[j];
    c = solution[k];
    d = solution[l];

    /*
     *  For a cargo to successfully load onto a vehicle, the vehicle
     *  and the cargo must be at the same node and the vehicle must
     *  still be at that node in the next time step.
     */
    if( C->value_attribute[i][a].type == NODE &&
        C->value_attribute[k][c].type == VEHICLE &&
        C->value_attribute[k][c].number ==
            C->variable_attribute[j].number &&
        !(C->value_attribute[i][a].number ==
                C->value_attribute[j][b].number &&
          C->value_attribute[i][a].number ==
                C->value_attribute[l][d].number) ) {
        return( 0 );
    }
    /*
     *  For a cargo to successfully unload from a vehicle, the vehicle
     *  and the cargo must be at the same node in the next time
     *  step and the vehicle must have been at that node in the
     *  previous time step.
     */
    else
    if( C->value_attribute[i][a].type == VEHICLE &&
        C->value_attribute[k][c].type == NODE &&
        C->value_attribute[i][a].number ==
            C->variable_attribute[j].number &&
        !(C->value_attribute[k][c].number ==
                C->value_attribute[j][b].number &&
          C->value_attribute[k][c].number ==
                C->value_attribute[l][d].number) ) {
        return( 0 );
    }
    else {
        return( 1 );
    }
}

static int
check_action_move_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j, k, l;
    int         a, b, c, d;

    i = C->scheme[r][0]; /* v_loc(t)    */
    j = C->scheme[r][1]; /* n_fuel(t)   */
    k = C->scheme[r][2]; /* v_loc(t+1)  */
    l = C->scheme[r][3]; /* n_fuel(t+1) */

    a = solution[i];
    b = solution[j];
    c = solution[k];
    d = solution[l];

    /*
     *  For a vehicle to successfully move, there must be at least
     *  one unit of fuel available at the origination node and the
     *  nodes must be adjacent.
     */
    if( C->value_attribute[i][a].type == NODE &&
        C->value_attribute[k][c].type == NODE &&
        C->value_attribute[i][a].number !=
            C->value_attribute[k][c].number &&
        C->value_attribute[i][a].number ==
            C->variable_attribute[j].number &&
        !(sp_node_distance( C->value_attribute[i][a].number,
                          C->value_attribute[k][c].number ) == 1 &&
          C->value_attribute[j][b].number >= 1 &&
          (C->value_attribute[j][b].number-1) ==
                C->value_attribute[l][d].number) ) {
        return( 0 );
    }
    else {
        return( 1 );
    }
}

static int
check_vehicle_capacity_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j, v, a;
    int         load;

    for( i = 1; i <= get_num_vehicles(); i++ ) {
        load = 0;
        for( j = 0; j < C->arity[r]; j++ ) {
            v = C->scheme[r][j];
            a = solution[v];
            if( C->value_attribute[v][a].type == VEHICLE &&
                C->value_attribute[v][a].number == i ) {
                load++;
            }
        }
        if( load > get_max_capacity( i ) ) {
            return( 0 );
        }
    }

    return( 1 );
}

static int
check_single_move_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j, k, l;
    int         a, b, c, d;

    i = C->scheme[r][0]; /* a.v_loc(t)   */
    j = C->scheme[r][1]; /* b.v_loc(t)   */
    k = C->scheme[r][2]; /* a.v_loc(t+1) */
    l = C->scheme[r][3]; /* b.v_loc(t+1) */

    a = solution[i];
    b = solution[j];
    c = solution[k];
    d = solution[l];

    if( C->value_attribute[i][a].number != C->value_attribute[k][c].number &&
        C->value_attribute[j][b].number != C->value_attribute[l][d].number &&
        C->value_attribute[i][a].number == C->value_attribute[j][b].number ) {
        return( 0 );
    }
    else {
        return( 1 );
    }
}

static int
check_action_squirt_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         g, h, i, j, k, l;
    int         a, b, c, d, e, f;

    g = C->scheme[r][0]; /* squirt_from(t)     */
    h = C->scheme[r][1]; /* fuel_at_nodeA(t)   */
    i = C->scheme[r][2]; /* fuel_at_nodeB(t)   */
    j = C->scheme[r][3]; /* squirt_to(t+1)     */
    k = C->scheme[r][4]; /* fuel_at_nodeA(t+1) */
    l = C->scheme[r][5]; /* fuel_at_nodeB(t+1) */

    a = solution[g];
    b = solution[h];
    c = solution[i];
    d = solution[j];
    e = solution[k];
    f = solution[l];

    /*
     *  For a squirt action to successfully transfer fuel from nodeA
     *  to nodeB, there must be at least two units of fuel at nodeA.
     */
    if( C->value_attribute[g][a].type == NODE &&
        C->value_attribute[j][d].type == NODE &&
        C->value_attribute[g][a].number !=
            C->value_attribute[j][d].number &&
        C->value_attribute[g][a].number ==
            C->variable_attribute[h].number &&
        C->value_attribute[j][d].number ==
            C->variable_attribute[i].number &&
        !(C->value_attribute[h][b].number >= 2 &&
          (C->value_attribute[h][b].number-1) ==
                C->value_attribute[k][e].number &&
          (C->value_attribute[i][c].number+1) ==
                C->value_attribute[l][f].number) ) {
        return( 0 );
    }
    /*
     *  The only way for a nodeB to get more fuel is through
     *  a squirt action from some other node.
     */
    else
    if( (C->value_attribute[i][c].number+1) ==
                C->value_attribute[l][f].number &&
        !(C->value_attribute[g][a].type == NODE &&
          C->value_attribute[j][d].type == NODE &&
          C->value_attribute[g][a].number !=
              C->value_attribute[j][d].number &&
          C->value_attribute[j][d].number ==
              C->variable_attribute[i].number) ) {
        return( 0 );
    }
    else {
        return( 1 );
    }
}

static int
check_single_squirt_move_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j, k, l;
    int         a, b, c, d;

    i = C->scheme[r][0]; /* squirt_from(t) */
    j = C->scheme[r][1]; /* v_loc(t)       */
    k = C->scheme[r][2]; /* squirt_to(t+1) */
    l = C->scheme[r][3]; /* v_loc(t+1)     */

    a = solution[i];
    b = solution[j];
    c = solution[k];
    d = solution[l];

    if( C->value_attribute[i][a].number != C->value_attribute[k][c].number &&
        C->value_attribute[j][b].number != C->value_attribute[l][d].number &&
        C->value_attribute[i][a].number == C->value_attribute[j][b].number ) {
        return( 0 );
    }
    else {
        return( 1 );
    }
}

