
/*
 *  CPlan solves planning problems formulated as constraint satisfaction
 *  problems.
 *
 *  Copyright (C) 1999  Peter van Beek and Xinguang Chen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.  See the GNU General Public License
 *  for more details (see the file called Copying or contact the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA).
 */

#include "mystery.h"
#include "../Share/global.h"
#include "../Share/csp.h"

/*
 *  Private routines.
 */
static void	set_domain();
static void	set_value_attributes();


/*
 *  Generate a CSP corresponding to a blocks world planning problem.
 */
CSP_type *
generate( time )
    int         time;
{
    CSP_type    *C;
    int         i, a, b, t, t1, t2, v;
    int         next_con;

    C = (CSP_type *)new_calloc( 1, sizeof(CSP_type) );

    C->time = time;
    if( !squirt_action() ) {
        C->n = (time+1) * (get_num_cargo() +
			   get_num_vehicles() +
			   get_num_nodes());
    }
    else {
        C->n = (time+1) * (get_num_cargo() +
			   get_num_vehicles() +
			   get_num_nodes() +
			   2);
    }

    /*
     *  Allocate space for n variables.
     */
    C->domain_size = new_calloc( C->n+1, sizeof(int) );
    C->variable_attribute = (VARA *)new_calloc( C->n+1, sizeof(VARA) );
    C->value_attribute = (VALA **)new_calloc( C->n+1, sizeof(VALA *) );

    /*
     *  Record the attributes of each variable.
     *  Initialize the domain of each variable.
     */
    for( t = 0; t <= time; t++ ) {
        /*
         * Cargo variables
         */
        for( i = 1; i <= get_num_cargo(); i++ ) {
            v = get_var( C, t, C_LOC, i );
            set_variable_attributes( C, v, C_LOC, t, get_cargo_name(i) );
            set_domain( C, v );
        }
        /*
         * Vehicle variables
         */
        for( i = 1; i <= get_num_vehicles(); i++ ) {
            v = get_var( C, t, V_LOC, i );
            set_variable_attributes( C, v, V_LOC, t, get_vehicle_name(i) );
            set_domain( C, v );
        }
        /*
         * Node variables
         */
        for( i = 1; i <= get_num_nodes(); i++ ) {
            v = get_var( C, t, N_FUEL, i );
            set_variable_attributes( C, v, N_FUEL, t, get_node_name(i) );
            set_domain( C, v );
        }
        /*
         * Squirt variables
         */
        if( squirt_action() ) {
            v = get_var( C, t, SQUIRT, 1 );
            set_variable_attributes( C, v, SQUIRT, t, 1 );
            set_domain( C, v );
            v = get_var( C, t, SQUIRT, 2 );
            set_variable_attributes( C, v, SQUIRT, t, 2 );
            set_domain( C, v );
        }
    }

    /*
     *  Is the variable type visible? (1) or hidden? (0)
     */
    C->visible[C_LOC]  = 1;
    C->visible[V_LOC]  = 0;
    C->visible[N_FUEL] = 0;
    C->visible[SQUIRT] = 0;
    C->n_visible = (time+1) * get_num_cargo();


    next_con = 0;

    /*
     *  Record the arity, scheme, and relation of
     *  binary distance constraints on cargo.
     */
    for( t1 = 0;    t1 <  time; t1++ )
    for( t2 = t1+1; t2 <= time; t2++ ) {
        for( a = 1; a <= get_num_cargo(); a++ ) {
            set_constraint( C, next_con, 2, DISTANCE_C_LOC_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t1, C_LOC, a );
            C->scheme[next_con][1] = get_var( C, t2, C_LOC, a );
            next_con++;
        }
    }

    /*
     *  Record the arity, scheme, and relation of
     *  binary distance constraints on vehicles.
     */
    for( t1 = 0;    t1 <  time; t1++ )
    for( t2 = t1+1; t2 <= time; t2++ ) {
        for( a = 1; a <= get_num_vehicles(); a++ ) {
            set_constraint( C, next_con, 2, DISTANCE_V_LOC_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t1, V_LOC, a );
            C->scheme[next_con][1] = get_var( C, t2, V_LOC, a );
            next_con++;
        }
    }

    /*
     *  Record the arity, scheme, and relation of
     *  binary distance constraints on fuel at nodes.
     */
    for( t1 = 0;    t1 <  time; t1++ )
    for( t2 = t1+1; t2 <= time; t2++ ) {
        for( a = 1; a <= get_num_nodes(); a++ ) {
            set_constraint( C, next_con, 2, DISTANCE_N_FUEL_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t1, N_FUEL, a );
            C->scheme[next_con][1] = get_var( C, t2, N_FUEL, a );
            next_con++;
        }
    }

    /*
     *  Record the arity, scheme, and relation of
     *  4-ary constraints on cargo and vehicles.
     */
    for( t = 0; t < time; t++ ) {
        for( a = 1; a <= get_num_cargo();    a++ )
        for( b = 1; b <= get_num_vehicles(); b++ ) {
            set_constraint( C, next_con, 4, ACTION_LOAD_UNLOAD_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t,   C_LOC, a );
            C->scheme[next_con][1] = get_var( C, t,   V_LOC, b );
            C->scheme[next_con][2] = get_var( C, t+1, C_LOC, a );
            C->scheme[next_con][3] = get_var( C, t+1, V_LOC, b );
            next_con++;
        }
    }

    /*
     *  Record the arity, scheme, and relation of
     *  4-ary constraints on vehicles and nodes.
     */
    for( t = 0; t < time; t++ ) {
        for( a = 1; a <= get_num_vehicles(); a++ )
        for( b = 1; b <= get_num_nodes();    b++ ) {
            set_constraint( C, next_con, 4, ACTION_MOVE_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t,   V_LOC,  a );
            C->scheme[next_con][1] = get_var( C, t,   N_FUEL, b );
            C->scheme[next_con][2] = get_var( C, t+1, V_LOC,  a );
            C->scheme[next_con][3] = get_var( C, t+1, N_FUEL, b );
            next_con++;
        }
    }

    /*
     *  Record the arity, scheme, and relation of
     *  varying arity constraints on cargo.
     */
    for( t = 0; t < time; t++ ) {
        set_constraint( C, next_con, get_num_cargo(), VEHICLE_CAPACITY_CONSTRAINT );
        for( a = 1; a <= get_num_cargo(); a++ ) {
            C->scheme[next_con][a-1] = get_var( C, t, C_LOC, a );
        }
        next_con++;
    }

    /*
     *  Record the arity, scheme, and relation of
     *  4-ary constraints on vehicles.
     */
    for( t = 0; t < time; t++ ) {
        for( a = 1;   a <= get_num_vehicles(); a++ )
        for( b = a+1; b <= get_num_vehicles(); b++ ) {
            set_constraint( C, next_con, 4, SINGLE_MOVE_CONSTRAINT );
            C->scheme[next_con][0] = get_var( C, t,   V_LOC, a );
            C->scheme[next_con][1] = get_var( C, t,   V_LOC, b );
            C->scheme[next_con][2] = get_var( C, t+1, V_LOC, a );
            C->scheme[next_con][3] = get_var( C, t+1, V_LOC, b );
            next_con++;
        }
    }

    /*
     *  Record the arity, scheme, and relation of
     *  6-ary constraints on squirt actions and nodes.
     */
    if( squirt_action() ) {
        for( t = 0; t < time; t++ ) {
            for( a = 1; a <= get_num_nodes(); a++ )
            for( b = 1; b <= get_num_nodes(); b++ ) if( a != b ) {
                set_constraint( C, next_con, 6, ACTION_SQUIRT_CONSTRAINT );
                C->scheme[next_con][0] = get_var( C, t,   SQUIRT, 1 );
                C->scheme[next_con][1] = get_var( C, t,   N_FUEL, a );
                C->scheme[next_con][2] = get_var( C, t,   N_FUEL, b );
                C->scheme[next_con][3] = get_var( C, t+1, SQUIRT, 2 );
                C->scheme[next_con][4] = get_var( C, t+1, N_FUEL, a );
                C->scheme[next_con][5] = get_var( C, t+1, N_FUEL, b );
                next_con++;
            }
        }
    }

    /*
     *  Record the arity, scheme, and relation of
     *  4-ary constraints on squirt actions and vehicles.
     */
    if( squirt_action() ) {
        for( t = 0; t < time; t++ ) {
            for( a = 1; a <= get_num_vehicles(); a++ ) {
                set_constraint( C, next_con, 4, SINGLE_SQUIRT_MOVE_CONSTRAINT );
                C->scheme[next_con][0] = get_var( C, t,   SQUIRT, 1 );
                C->scheme[next_con][1] = get_var( C, t,   V_LOC,  a );
                C->scheme[next_con][2] = get_var( C, t+1, SQUIRT, 2 );
                C->scheme[next_con][3] = get_var( C, t+1, V_LOC,  a );
                next_con++;
            }
        }
    }

    C->m = next_con;

    /*
     *  When should the constraint be propagated?
     *
     *  C->checkable[r] = u means that constraint type r is propagated
     *          only when the number of uninstantiated variables in the
     *          scheme of the constraint is less than or equal to u.
     *          Constraint propagation then potentially reduces the
     *          domains of each of the uninstantiated variables.
     *  
     *  In particular,
     *    == 1    means only forward checking done on this constraint
     *    >= 2    means arc consistency checking done on this constraint
     *    == HUGE means constraint is always propagated
     */
    C->checkable[DISTANCE_C_LOC_CONSTRAINT] = HUGE;
    C->checkable[DISTANCE_V_LOC_CONSTRAINT] = HUGE;
    C->checkable[DISTANCE_N_FUEL_CONSTRAINT] = HUGE;
    C->checkable[ACTION_LOAD_UNLOAD_CONSTRAINT] = HUGE;
    C->checkable[ACTION_MOVE_CONSTRAINT] = HUGE;
    C->checkable[VEHICLE_CAPACITY_CONSTRAINT] = HUGE;
    C->checkable[SINGLE_MOVE_CONSTRAINT] = HUGE;
    C->checkable[ACTION_SQUIRT_CONSTRAINT] = HUGE;
    C->checkable[SINGLE_SQUIRT_MOVE_CONSTRAINT] = HUGE;

    construct_index( C );

    return( C );
}

int
get_var( C, state, type, number )
    CSP_type    *C;
    int         state, type, number;
{
    int         i;

    if( !squirt_action() ) {
        i = state * (get_num_cargo() +
		     get_num_vehicles() +
		     get_num_nodes());
    }
    else {
        i = state * (get_num_cargo() +
		     get_num_vehicles() +
		     get_num_nodes() +
		     2);
    }

    if( type == C_LOC ) {
        i += number;
    }
    else
    if( type == V_LOC ) {
        i += get_num_cargo() + number;
    }
    else
    if( type == N_FUEL ) {
        i += get_num_cargo() + get_num_vehicles() + number;
    }
    else
    if( type == SQUIRT ) {
        i += get_num_cargo() + get_num_vehicles() + get_num_nodes() + number;
    }

    return( i );
}

static void
set_value_attributes( C, v, a, type, number )
    CSP_type    *C;
    int         type, number;
{
    C->value_attribute[v][a].type = type;
    C->value_attribute[v][a].number = number;
}

/*
 *  Initialize the domain of a variable.
 *  Record the attributes of each value.
 */
static void
set_domain( C, v )
    CSP_type    *C;
    int         v;
{
    int         state, type, i, j;
    int         next_val;

    get_variable_attributes( C, v, &type, &state, &i );
    if( type == C_LOC ) {
        if( state == 0 ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
            set_value_attributes( C, v, 0, NODE, get_init_cargo(i) );
            C->domain_size[v] = 1;
        }
        else
        if( state == C->time ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
            set_value_attributes( C, v, 0, NODE, get_goal_cargo(i) );
            C->domain_size[v] = 1;
        }
        else {
            C->value_attribute[v] = (VALA *)new_calloc(
                        get_num_vehicles() + get_num_nodes(), sizeof(VALA) );
            next_val = 0;
            for( j = 1; j <= get_num_vehicles(); j++ ) {
                set_value_attributes( C, v, next_val, VEHICLE, j );
                next_val++;
            }
            for( j = 1; j <= get_num_nodes(); j++ ) {
                set_value_attributes( C, v, next_val, NODE, j );
                next_val++;
            }
            C->domain_size[v] = next_val;
        }
    }
    else
    if( type == V_LOC ) {
        if( state == 0 ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
            set_value_attributes( C, v, 0, NODE, get_init_vehicle(i) );
            C->domain_size[v] = 1;
        }
        else {
            C->domain_size[v] = get_num_nodes();
            C->value_attribute[v] = (VALA *)new_calloc(
                        C->domain_size[v], sizeof(VALA) );
            next_val = 0;
            for( j = 1; j <= get_num_nodes(); j++ ) {
                set_value_attributes( C, v, next_val, NODE, j );
                next_val++;
            }
        }
    }
    else
    if( type == N_FUEL ) {
        if( state == 0 ) {
            C->value_attribute[v] = (VALA *)new_calloc( 1, sizeof(VALA) );
            set_value_attributes( C, v, 0, NUMBER, get_max_fuel(i) );
            C->domain_size[v] = 1;
        }
        else {
            C->domain_size[v] = get_max_fuel(i)+1;
            if( (C->domain_size[v] == 1) && squirt_action() ) {
                C->domain_size[v]++;
            }
            C->value_attribute[v] = (VALA *)new_calloc(
                        C->domain_size[v], sizeof(VALA) );
            next_val = 0;
            for( j = C->domain_size[v]-1; j >= 0; j-- ) {
                set_value_attributes( C, v, next_val, NUMBER, j );
                next_val++;
            }
        }
    }
    else
    if( type == SQUIRT ) {
        C->domain_size[v] = get_num_nodes();
        C->value_attribute[v] = (VALA *)new_calloc(
                    C->domain_size[v], sizeof(VALA) );
        next_val = 0;
        for( j = 1; j <= get_num_nodes(); j++ ) {
            set_value_attributes( C, v, next_val, NODE, j );
            next_val++;
        }
    }
}

